/*++

Copyright (c) 1998 Microsoft Corporation

Module Name:

    expool.c

Abstract:

    Hack Executive Pool Stuff

Author:

    MSKK Oem Tools Project    07-Sep-1998

Revision History:

--*/

#include "precomp.h"
#pragma hdrstop

HKAPLUGIN hEXPOOL;

HKATYPE EXPOOL_Type_PoolHeader = 0;         // PoolBlock\\PoolHeader
HKATYPE EXPOOL_Anomaly_PoolHeader = 0;      // AnomalyDescriptor\\PoolBlock\\PoolHeader
HKATYPE EXPOOL_Type_Paged = 0;              // PoolBlock\\Paged
HKATYPE EXPOOL_Type_NonPaged = 0;           // PoolBlock\\NonPaged
HKATYPE EXPOOL_Type_PagedFree = 0;          // PoolBlock\\Paged\\FreedBlock
HKATYPE EXPOOL_Type_NonPagedFree = 0;       // PoolBlock\\NonPaged\\FreedBlock

HKATYPE EXPOOL_Type_PoolDescriptor = 0;     // PoolControl\\PoolDescriptor
HKATYPE EXPOOL_Anomaly_PoolDescriptor = 0;  // AnomalyDescriptor\\PoolControl\\PoolDescriptor

HKATYPE EXPOOL_Type_MMPOOL_PagedPool = 0;
HKATYPE EXPOOL_Type_MMPOOL_NonPagedPool = 0;

HKATYPE EXPOOL_Anomaly_InaccessibleMemory = 0; // AnomalyDescriptor\\InaccessibleMemory

#define PACK_POOL_INDEX(Index) ((UCHAR)(((Index) << 4) | (Index)))
#define UNPACK_POOL_INDEX(Index) ((ULONG)((Index) & 0xf))

#define EXPOOL_FORMAT_BUFFER_SIZE 1024

UINT_PTR
CALLBACK
EXPOOL_KanalyzeInterface(
    IN HKAPLUGIN PlugInHandle,
    IN UINT32 OperationCode,
    IN OUT UINT_PTR Param1,
    IN OUT PVOID Param2
    );

UINT32
EXPOOL_Initialize(
    IN HKAPLUGIN h,
    IN UINT_PTR Pass
    );

UINT32
EXPOOL_RegisterTypes(
    IN HKAPLUGIN h,
    IN PKANALYZE_REGISTER_TYPES_PROC fnRegister
    );

UINT32
EXPOOL_LocateItems(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fn
    );

UINT32
EXPOOL_ProcessItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fn
    );

UINT32
EXPOOL_NotifyItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fn
    );

UINT32
EXPOOL_Terminate(
    IN HKAPLUGIN h,
    IN KanalyzeFinalOutcome Outcome
    );

UINT32
EXPOOL_FormatDescription(
    IN HKAPLUGIN h,
    IN PKANALYZE_DATA_ITEM_STRING_REQUEST Request,
    OUT PUINT32 CharsWritten
    );

BOOL
EXPOOL_IsThisPageUsedAsPoolBlockPage(
    IN HKAPLUGIN h,
    IN POOL_TYPE PoolType,
    IN PUCHAR PoolPage
    );

UINT32
EXPOOL_LocatePoolBlocks(
    HKAPLUGIN h,
    KA_ITEM_ID id,
    PKANALYZE_LOCATE_PROC fnLocate,
    POOL_TYPE PoolType,
    UINT_PTR Address,
    SIZE_T Size
    );

UINT32
EXPOOL_LocatePoolDescriptorAll(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    );

VOID
EXPOOL_FormatPoolHeader(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
EXPOOL_FormatPoolBlock(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
EXPOOL_FormatPoolDescriptor(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
EXPOOL_FormatAnomalyPoolHeader(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
EXPOOL_FormatAnomalyPoolDescriptor(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

LPWSTR
EXPOOL_TagToString(
    ULONG32 PoolTag
    );

UINT32
EXPOOL_PoolHeader_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    );

UINT32
EXPOOL_PoolBlock_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    );

UINT32
EXPOOL_PoolDescriptor_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    );



BOOL
EXPOOL_RegisterPlugIn(
    IN KANALYZE_REGISTER_PLUGIN_PROC RegistrationRoutine,
    IN ULONG32 CrashPlatform,
    IN ULONG32 CrashBuildNumber,
    PVOID Reserved1,
    UINT_PTR Reserved2
    )
{

    hEXPOOL = (RegistrationRoutine)(
                  KANALYZE_INTERFACE_VERSION,
                  PLUGIN_EXPOOL,
                  EXPOOL_KanalyzeInterface,
                  &KanalyzeExtensionApis,
                  0,
                  0
                  );

    if (!hEXPOOL) {
        return FALSE;
    }

    return TRUE;
}



UINT_PTR
CALLBACK
EXPOOL_KanalyzeInterface(
    IN HKAPLUGIN PlugInHandle,
    IN UINT32 OperationCode,
    IN OUT UINT_PTR Param1,
    IN OUT PVOID Param2
    )

/*++

Routine Description:

    This is an interface routine between this plugin and the engine. The engine
    calls this routine.


Arguments:

    see the interface document

Return Value:

    Win32 Error Code

--*/

{
    UINT_PTR rc;

    rc = 0;

    switch (OperationCode) {
    case KA_INITIALIZE:
            rc = EXPOOL_Initialize(PlugInHandle, Param1);
            break;
    case KA_REGISTER_TYPES:
            rc = EXPOOL_RegisterTypes(PlugInHandle, (PKANALYZE_REGISTER_TYPES_PROC)Param2);
            break;
    case KA_START_LOCATE_ITEMS:
            rc = EXPOOL_LocateItems(PlugInHandle, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_PROCESS_ITEM:
            rc = EXPOOL_ProcessItem(PlugInHandle, (KA_ITEM_ID)Param1, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_NOTIFY_ITEM:
            rc = EXPOOL_NotifyItem(PlugInHandle, (KA_ITEM_ID)Param1, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_PROCESS_RESULT:
            rc = NO_ERROR; // not handle this code
            break;
    case KA_TERMINATE:
            rc = EXPOOL_Terminate(PlugInHandle, (KanalyzeFinalOutcome)Param1);
            break;
    case KA_FORMAT_DATA_ITEM_DESCRIPTION:
            rc = EXPOOL_FormatDescription(PlugInHandle, (PKANALYZE_DATA_ITEM_STRING_REQUEST)Param1, (PUINT32)Param2);
            break;
    default:
            rc = ERROR_INVALID_FUNCTION;
    }

    return rc;
}



UINT32
EXPOOL_Initialize(
    IN HKAPLUGIN h,
    IN UINT_PTR Pass
    )
{
    return NO_ERROR;
}


UINT32
EXPOOL_RegisterTypes(
    IN HKAPLUGIN h,
    IN PKANALYZE_REGISTER_TYPES_PROC fnRegister
    )
{
    UINT32 rc;
    HKATYPE AllowedContainerType[2];

    rc = NO_ERROR;


    AllowedContainerType[0] = KxGetTypeHandle(h, L"MmPool\\PagedPool");
    AllowedContainerType[1] = KxGetTypeHandle(h, L"MmPool\\NonPagedPool");
    rc = (fnRegister)(
             h,
             L"PoolBlock\\PoolHeader",
             L"Pool Header",
             2,
             &AllowedContainerType[0],
             0,
             SupportFull,
             EXPOOL_PoolHeader_Method,
             &EXPOOL_Type_PoolHeader
             );
    if (rc != NO_ERROR)
        return rc;

    rc = (fnRegister)(
             h,
             L"AnomalyDescriptor\\PoolBlock\\PoolHeader",
             L"Anomaly Descriptor of Pool Header",
             0,
             NULL,
             KA_TC_DONT_CHECK_OVERLAP,
             SupportFull,
             NULL,
             &EXPOOL_Anomaly_PoolHeader
             );
    if (rc != NO_ERROR)
        return rc;

    AllowedContainerType[0] = KxGetTypeHandle(h, L"MmPool\\PagedPool");
    rc = (fnRegister)(
             h,
             L"PoolBlock\\Paged",
             L"Paged Pool Block In Use",
             1,
             &AllowedContainerType[0],
             0,
             SupportFull,
             EXPOOL_PoolBlock_Method,
             &EXPOOL_Type_Paged
             );
    if (rc != NO_ERROR)
        return rc;

    AllowedContainerType[0] = KxGetTypeHandle(h, L"MmPool\\NonPagedPool");
    rc = (fnRegister)(
             h,
             L"PoolBlock\\NonPaged",
             L"Non Paged Pool Block In Use",
             1,
             &AllowedContainerType[0],
             KA_TC_LOCATE_ALL_ITEMS,   //can locate all item
             SupportFull,
             EXPOOL_PoolBlock_Method,
             &EXPOOL_Type_NonPaged
             );
    if (rc != NO_ERROR)
        return rc;

    rc = (fnRegister)(
             h,
             L"PoolBlock\\Paged\\FreedBlock",
             L"Pool Block Not Used",
             0,
             NULL,
             0,
             SupportFull,
             EXPOOL_PoolBlock_Method,
             &EXPOOL_Type_PagedFree
             );
    if (rc != NO_ERROR)
        return rc;

    rc = (fnRegister)(
             h,
             L"PoolBlock\\NonPaged\\FreedBlock",
             L"Pool Block Not Used",
             0,
             NULL,
             0,
             SupportFull,
             EXPOOL_PoolBlock_Method,
             &EXPOOL_Type_NonPagedFree
             );
    if (rc != NO_ERROR)
        return rc;

    rc = (fnRegister)(
             h,
             L"PoolControl\\PoolDescriptor",
             L"PoolDescriptor",
             0,
             NULL,
             0,
             SupportFull,
             EXPOOL_PoolDescriptor_Method,
             &EXPOOL_Type_PoolDescriptor
             );
    if (rc != NO_ERROR)
        return rc;

    rc = (fnRegister)(
             h,
             L"AnomalyDescriptor\\PoolControl\\PoolDescriptor",
             L"Anomaly Descriptor of PoolDescriptor",
             0,
             NULL,
             KA_TC_DONT_CHECK_OVERLAP,
             SupportFull,
             NULL,
             &EXPOOL_Anomaly_PoolDescriptor
             );
    if (rc != NO_ERROR)
        return rc;


    rc = (fnRegister)(
             h,
             L"MmPool\\PagedPool",
             NULL,
             0,
             NULL,
             0,
             SupportDemandNotification,
             NULL,
             &EXPOOL_Type_MMPOOL_PagedPool
             );
    if (rc != NO_ERROR)
        return rc;

    rc = (fnRegister)(
             h,
             L"MmPool\\NonPagedPool",
             NULL,
             0,
             NULL,
             0,
             SupportDemandNotification,
             NULL,
             &EXPOOL_Type_MMPOOL_NonPagedPool
             );
    if (rc != NO_ERROR)
        return rc;

    EXPOOL_Anomaly_InaccessibleMemory = KxGetTypeHandle(
                                            h,
                                            L"AnomalyDescriptor\\InaccessibleMemory"
                                            );

    return rc;
}



UINT32
EXPOOL_LocateItems(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;

    rc = EXPOOL_LocatePoolDescriptorAll(h, fnLocate);

    return rc;
}



UINT32
EXPOOL_ProcessItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    return NO_ERROR;
}



UINT32
EXPOOL_NotifyItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;
    KA_DATA_ITEM_INFO ItemInfo;
    POOL_TYPE PoolType;

    rc = KxQueryDataItem(h, id, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    if (ItemInfo.TypeHandle != EXPOOL_Type_MMPOOL_PagedPool
     && ItemInfo.TypeHandle != EXPOOL_Type_MMPOOL_NonPagedPool) {
        // do nothing
        return NO_ERROR;
    }

    if (ItemInfo.TypeHandle == EXPOOL_Type_MMPOOL_PagedPool) {
        PoolType = PagedPool;
    } else {
        PoolType = NonPagedPool;
    }

    rc = EXPOOL_LocatePoolBlocks(h, id, fnLocate, PoolType, ItemInfo.Address, ItemInfo.Size);

    return rc;
}



UINT32
EXPOOL_Terminate(
    IN HKAPLUGIN h,
    IN KanalyzeFinalOutcome Outcome
    )
{
    return NO_ERROR;
}



UINT32
EXPOOL_FormatDescription(
    IN HKAPLUGIN h,
    IN PKANALYZE_DATA_ITEM_STRING_REQUEST Request,
    OUT PUINT32 CharsWritten
    )
{
    UINT32 rc;
    WCHAR Buffer[EXPOOL_FORMAT_BUFFER_SIZE];
    KA_DATA_ITEM_INFO ItemInfo;
    UINT32 CharsNeeds;
    KA_ITEM_ID id;

    *CharsWritten = 0;

    id = Request->DataItemId;

    rc = KxQueryDataItem(h, id, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    if (ItemInfo.TypeHandle == EXPOOL_Anomaly_PoolHeader) {

        EXPOOL_FormatAnomalyPoolHeader(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == EXPOOL_Anomaly_PoolDescriptor) {

        EXPOOL_FormatAnomalyPoolDescriptor(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == EXPOOL_Type_PoolHeader) {

        EXPOOL_FormatPoolHeader(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == EXPOOL_Type_Paged ||
               ItemInfo.TypeHandle == EXPOOL_Type_NonPaged ||
               ItemInfo.TypeHandle == EXPOOL_Type_PagedFree ||
               ItemInfo.TypeHandle == EXPOOL_Type_NonPagedFree) {

        EXPOOL_FormatPoolBlock(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == EXPOOL_Type_PoolDescriptor) {

        EXPOOL_FormatPoolDescriptor(h, id, &ItemInfo, Buffer);

    } else {
        lstrcpy(Buffer, L"EXPOOL:\n");
    }

    CharsNeeds = lstrlen(Buffer) + 1;

    if (Request->MessageBufferSize < CharsNeeds) {
        *CharsWritten = CharsNeeds;
        return ERROR_MORE_DATA;
    } else {
        lstrcpy(Request->MessageBuffer, Buffer);
        *CharsWritten = CharsNeeds - 1;
        return NO_ERROR;
    }
}



///////////////////////////////////////////////////////////////////////////

BOOL
EXPOOL_IsThisPageUsedAsPoolBlockPage(
    IN HKAPLUGIN h,
    IN POOL_TYPE PoolType,
    IN PUCHAR PoolPage
    )
{
    PPOOL_HEADER Header, Header2;
    ULONG PoolIndex;
    ULONG NumberOfPagedPools;

    Header = (PPOOL_HEADER)PoolPage;
    NumberOfPagedPools = GetValue(h, L"ExpNumberOfPagedPools");

    do {
        //
        // sanity check in order to determine this page is used as pool.
        //
        if (Header->PreviousSize != 0) break;

        if (((Header->PoolIndex & 0xf0) >> 4) != (Header->PoolIndex & 0x0f)) break;

        PoolIndex = (ULONG)UNPACK_POOL_INDEX(Header->PoolIndex);
        if (PoolType == NonPagedPool)
            if (PoolIndex != 0) break;
        else
            if (PoolIndex > NumberOfPagedPools) break;

        if (Header->BlockSize == 0) break;
        if ((Header->BlockSize << POOL_BLOCK_SHIFT) > POOL_PAGE_SIZE) break;

        Header2 = (PPOOL_HEADER)(PoolPage + (Header->BlockSize << POOL_BLOCK_SHIFT));

        if (Header2->PreviousSize != Header->BlockSize) break;
        if (((Header2->PoolIndex & 0xf0) >> 4) != (Header2->PoolIndex & 0x0f)) break;

        PoolIndex = (ULONG)UNPACK_POOL_INDEX(Header2->PoolIndex);
        if (PoolType == NonPagedPool)
            if (PoolIndex != 0) break;
        else
            if (PoolIndex > NumberOfPagedPools) break;

        if (Header2->BlockSize == 0) break;
        if ((Header2->BlockSize << POOL_BLOCK_SHIFT) > POOL_PAGE_SIZE) break;

        return TRUE;

    } while (0);

    return FALSE;
}



UINT32
EXPOOL_LocatePoolBlocks(
    HKAPLUGIN h,
    KA_ITEM_ID id,
    PKANALYZE_LOCATE_PROC fnLocate,
    POOL_TYPE PoolType,
    UINT_PTR Address,
    SIZE_T Size
    )
{
    UINT32 rc;
    KANALYZE_DATA_ITEM DataItem;
    KA_ITEM_ID ItemId, HeaderId, PreviousHeaderId;
    HKATYPE TypeHandle;
    ULONG Offset, PrevSize;
    SIZE_T BytesRead;
    PPOOL_HEADER Header;
    UCHAR PageBuffer[PAGE_SIZE];
    ULONG NumberOfPagedPools;
    BOOL fBadBlockSize, fBadPrevSize, fBadPoolIndex, fFreedBlock;
    ULONG PoolIndex;
    SIZE_T PoolPageSize;

    NumberOfPagedPools = GetValue(h, L"ExpNumberOfPagedPools");

    if (PoolType == NonPagedPool) {
        TypeHandle = EXPOOL_Type_NonPaged;
    } else {
        TypeHandle = EXPOOL_Type_Paged;
    }

    if (Size > PAGE_SIZE) {
        //
        // This block is large. It doesn't have pool blocks.
        //
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = TypeHandle;
        DataItem.Address = Address;
        DataItem.Size = Size;
        DataItem.TypeSpecificParam = 0;
        rc = (fnLocate)(h, id, &DataItem, 0, &ItemId);
        return rc;
    }

    // this page is one page.

    rc = KxReadMemory(h, NULL, Address, PageBuffer, PAGE_SIZE, &BytesRead);
    if (rc != NO_ERROR) {
        if (PoolType == PagedPool) {
            return NO_ERROR;
        }
        return ERROR_NOACCESS;
    }

    if (!EXPOOL_IsThisPageUsedAsPoolBlockPage(h, PoolType, PageBuffer)) {
        //
        // This page is used for one page.
        //
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = TypeHandle;
        DataItem.Address = Address;
        DataItem.Size = Size;
        DataItem.TypeSpecificParam = 0;
        rc = (fnLocate)(h, id, &DataItem, 0, &ItemId);
        return rc;
    }

    Offset = 0;
    PrevSize = 0;
    PreviousHeaderId = id;
#if defined(_ALPHA_)
    PoolPageSize = PAGE_SIZE - POOL_SMALLEST_BLOCK; // alpha doesn't use last block.
#elif defined(_X86_)
    PoolPageSize = PAGE_SIZE;
#else
#error "Target platform not defined"
#endif

    while (Offset < PoolPageSize) {
        fBadBlockSize = FALSE;
        fBadPrevSize  = FALSE;
        fBadPoolIndex = FALSE;
        fFreedBlock = FALSE;

        Header = (PPOOL_HEADER)&PageBuffer[Offset];

        //
        // First, locate PoolHeader. PoolHeader will be connected to the previous
        // pool header item. The first header is connected to this page item.
        //
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = EXPOOL_Type_PoolHeader;
        DataItem.Address = Address + Offset;
        DataItem.Size = sizeof(POOL_HEADER);
        rc = (fnLocate)(h, PreviousHeaderId, &DataItem, 0, &HeaderId);
        if (rc != NO_ERROR) {
            return rc;
        }

        //
        // Next, check the PoolHeader
        //
        if (Header->BlockSize == 0 || (Header->BlockSize << POOL_BLOCK_SHIFT) > POOL_PAGE_SIZE) {
            fBadBlockSize = TRUE;
        }
        if (Header->PreviousSize != PrevSize) {
            fBadPrevSize = TRUE;
        }
        if ((Header->PoolIndex & 0x0f) != ((Header->PoolIndex >> 4) & 0x0f)) {
            fBadPoolIndex = TRUE;
        } else {
            PoolIndex = (ULONG)UNPACK_POOL_INDEX(Header->PoolIndex);
            if (PoolIndex != 0) {
                if (PoolType != PagedPool || PoolIndex > NumberOfPagedPools) {
                    fBadPoolIndex = TRUE;
                }
            }
        }
        //
        // BUGBUG: I don't have a good idea for checking PoolType for now.
        //
        if (Header->PoolType == 0) {
            fFreedBlock = TRUE;
        }

        if (fBadBlockSize || fBadPrevSize || fBadPoolIndex) {
            ANOMALY_POOLHEADER_EXTRA Extra;
            //
            // Something invalid found in PoolHeader, so locate an anomaly item
            // This item will be connected to pool header.
            //
            ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
            DataItem.TypeHandle = EXPOOL_Anomaly_PoolHeader;
            DataItem.Address = Address + Offset;
            DataItem.Size = sizeof(POOL_HEADER);
            DataItem.TypeSpecificParam = HeaderId;
            Extra.Reason = 0;
            if (fBadBlockSize) Extra.Reason |= ANOMALY_POOLHEADER_BLOCKSIZE;
            if (fBadPrevSize)  Extra.Reason |= ANOMALY_POOLHEADER_PREVSIZE;
            if (fBadPoolIndex) Extra.Reason |= ANOMALY_POOLHEADER_POOLINDEX;
            DataItem.TypeSpecificDataSize = sizeof(ANOMALY_POOLHEADER_EXTRA);
            DataItem.TypeSpecificData = (PVOID)&Extra;
            rc = (fnLocate)(h, HeaderId, &DataItem, 0, &ItemId);
            if (rc != NO_ERROR) {
                return rc;
            }
            //
            // BUGBUG : locate the rest of pool blocks as one block.
            // We can distinguish this from normal block by checking TypeSpecificParam
            //
            ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
            DataItem.TypeHandle = TypeHandle;
            DataItem.Address = Address + Offset + sizeof(POOL_HEADER);
            DataItem.Size = PAGE_SIZE - Offset - sizeof(POOL_HEADER);
            DataItem.TypeSpecificParam = HeaderId;
            rc = (fnLocate)(h, HeaderId, &DataItem, 0, &ItemId);
            if (rc != NO_ERROR) {
                return rc;
            }
            break;
        }

        //
        // Finally, locate the body of this pool block, connect it to pool header
        //
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        if (fFreedBlock) {
            if (TypeHandle == EXPOOL_Type_NonPaged)
                DataItem.TypeHandle = EXPOOL_Type_NonPagedFree;
            else
                DataItem.TypeHandle = EXPOOL_Type_PagedFree;
        } else {
            DataItem.TypeHandle = TypeHandle;
        }
        DataItem.Address = Address + Offset + sizeof(POOL_HEADER);
        DataItem.Size = (Header->BlockSize << POOL_BLOCK_SHIFT) - sizeof(POOL_HEADER);
        DataItem.TypeSpecificParam = HeaderId;
        rc = (fnLocate)(h, HeaderId, &DataItem, 0, &ItemId);
        if (rc != NO_ERROR) {
            return rc;
        }

        PrevSize = Header->BlockSize;
        Offset += (PrevSize << POOL_BLOCK_SHIFT);

        PreviousHeaderId = HeaderId;
    }

    return NO_ERROR;
}



UINT32
EXPOOL_LocatePoolDescriptor(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate,
    IN UINT_PTR AddressOfPoolDesc,
    IN POOL_TYPE PoolType
    )
{
    UINT32 rc;
    SIZE_T Address;
    SIZE_T Size;
    ULONG32 BytesRead;
    POOL_DESCRIPTOR PoolDesc;
    KANALYZE_DATA_ITEM DataItem;
    KA_ITEM_ID PoolDescId;
    KA_ITEM_ID DataItemId;
    ULONG i;
    WCHAR ListHeadTypeName[MAX_TYPE_NAME];
    HKATYPE hListHead;
    KA_ITEM_ID TypeSpecificData[POOL_LIST_HEADS];

    Address = AddressOfPoolDesc;
    Size = sizeof(POOL_DESCRIPTOR);

    ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
    DataItem.TypeHandle = EXPOOL_Type_PoolDescriptor;
    DataItem.Address = Address;
    DataItem.Size = Size;
    rc = (fnLocate)(h, 0, &DataItem, 0, &PoolDescId);
    if (rc != NO_ERROR) return rc;

    rc = KxReadMemory(h, NULL, Address, &PoolDesc, Size, &BytesRead);
    if (rc != NO_ERROR) {
        return ERROR_NOACCESS;
    }

    lstrcpy(ListHeadTypeName, L"ListEntry\\ListHead\\PoolDescriptor");
    hListHead = KxGetTypeHandle(h, ListHeadTypeName);

    ZeroMemory((PVOID)TypeSpecificData, sizeof(KA_ITEM_ID) * POOL_LIST_HEADS);

    for (i = 0; i < POOL_LIST_HEADS; i++) {
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = hListHead;
        DataItem.Address = AddressOfPoolDesc + FIELD_OFFSET(POOL_DESCRIPTOR, ListHeads[i]);
        DataItem.Size = sizeof(LIST_ENTRY);
        rc = (fnLocate)(h, PoolDescId, &DataItem, 0, &DataItemId);
        if (rc != NO_ERROR) return rc;

        TypeSpecificData[i] = DataItemId;
    }

    //
    // finally, locate again in order to update type-specific data
    //
    Address = AddressOfPoolDesc;
    Size = sizeof(POOL_DESCRIPTOR);
    ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
    DataItem.TypeHandle = EXPOOL_Type_PoolDescriptor;
    DataItem.Address = Address;
    DataItem.Size = Size;
    DataItem.TypeSpecificDataSize = sizeof(TypeSpecificData);
    DataItem.TypeSpecificData = (PVOID)TypeSpecificData;
    rc = (fnLocate)(h, 0, &DataItem, 0, &PoolDescId);

    return rc;
}



UINT32
EXPOOL_LocatePoolDescriptorAll(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;
    UINT_PTR Address;
    ULONG NumberOfPagedPools, i;

    do {
        Address = KxSymbolToAddress(h, L"NonPagedPoolDescriptor");
        rc = EXPOOL_LocatePoolDescriptor(h, fnLocate, Address, NonPagedPool);
        if (rc != NO_ERROR) break;

        Address = KxSymbolToAddress(h, L"NonPagedPoolDescriptorMS");
        rc = EXPOOL_LocatePoolDescriptor(h, fnLocate, Address, NonPagedPool);
        if (rc != NO_ERROR) break;


        NumberOfPagedPools = GetValue(h, L"ExpNumberOfPagedPools");
        Address = GetValue(h, L"ExpPagedPoolDescriptor");
        for (i = 0; i <= NumberOfPagedPools; i++) {
            rc = EXPOOL_LocatePoolDescriptor(h, fnLocate, Address, PagedPool);
            if (rc != NO_ERROR) break;

            Address += sizeof(POOL_DESCRIPTOR);
        }

    } while (0);

    return rc;
}



VOID
EXPOOL_FormatPoolHeader(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    UINT32 rc;
    LPWSTR s;
    POOL_HEADER PoolHeader;
    SIZE_T BytesRead;
    WCHAR type_str[64];

    rc = KxReadMemory(h, NULL, ItemInfo->Address, &PoolHeader, sizeof(POOL_HEADER), &BytesRead);
    if (rc != NO_ERROR) {
        s = RetrieveMessage(
                MSG_EXPOOL_POOLHEADER_UNREADABLE,
                FALSE,
                ItemInfo->Address
            );
        lstrcpy(Buffer, s);
        FreeMessage(s);
        return;
    }


    if (PoolHeader.PoolType == 0) {
        lstrcpy(type_str, L"Freed pool");
    } else {
        if ((PoolHeader.PoolType - 1) & BASE_POOL_TYPE_MASK) {
            lstrcpy(type_str, L"PagedPool");
        } else {
            lstrcpy(type_str, L"NonPagedPool");
        }
        if ((PoolHeader.PoolType - 1) & MUST_SUCCEED_POOL_TYPE_MASK) {
            lstrcat(type_str, L" MustSucceed");
        }
        if ((PoolHeader.PoolType - 1) & CACHE_ALIGNED_POOL_TYPE_MASK) {
            lstrcat(type_str, L" CacheAligned");
        }
        if ((PoolHeader.PoolType - 1) & POOL_QUOTA_MASK) {
            lstrcat(type_str, L" Quota");
        }
    }
    s = RetrieveMessage(
            MSG_EXPOOL_POOLHEADER,
            FALSE,
            ItemInfo->Address,
            type_str,
            PoolHeader.BlockSize << POOL_BLOCK_SHIFT,
            PoolHeader.PreviousSize << POOL_BLOCK_SHIFT,
            PoolHeader.PoolTag, EXPOOL_TagToString(PoolHeader.PoolTag)
            );

    ASSERT(lstrlen(s) < EXPOOL_FORMAT_BUFFER_SIZE);
    lstrcpy(Buffer, s);
    FreeMessage(s);
}


VOID
EXPOOL_FormatPoolBlock(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    LPWSTR s;
    KA_ITEM_ID HeaderId;

    if (ItemInfo->TypeSpecificParam == 0) {
        ASSERT(ItemInfo->Size > PAGE_SIZE);
        s = RetrieveMessage(
                MSG_EXPOOL_POOLBLOCK_BIG,
                FALSE,
                ItemInfo->Address,
                ItemInfo->Size / PAGE_SIZE
                );

        ASSERT(lstrlen(s) < EXPOOL_FORMAT_BUFFER_SIZE);
        lstrcpy(Buffer, s);
        FreeMessage(s);
    } else {
        HeaderId = ItemInfo->TypeSpecificParam;

        s = RetrieveMessage(
                MSG_EXPOOL_POOLBLOCK,
                FALSE,
                ItemInfo->Address,
                ItemInfo->Size,
                ItemInfo->TypeSpecificParam
                );

        ASSERT(lstrlen(s) < EXPOOL_FORMAT_BUFFER_SIZE);
        lstrcpy(Buffer, s);
        FreeMessage(s);
    }
}


VOID
EXPOOL_FormatAnomalyPoolHeader(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    LPWSTR s;
    ANOMALY_POOLHEADER_EXTRA Extra;
    SIZE_T BytesRead;
    WCHAR reason[64];
    UINT32 rc;

    rc = KxRetrieveTypeSpecificData(h, id, sizeof(Extra), &Extra, &BytesRead);
    if (rc != NO_ERROR) {
        lstrcpy(reason, L"can't get type-specific data");
    } else {
        reason[0] = UNICODE_NULL;
        if (Extra.Reason & ANOMALY_POOLHEADER_BLOCKSIZE)
            lstrcat(reason, L"BlockSize ");
        if (Extra.Reason & ANOMALY_POOLHEADER_PREVSIZE)
            lstrcat(reason, L"PreviousSize ");
        if (Extra.Reason & ANOMALY_POOLHEADER_POOLINDEX)
            lstrcat(reason, L"PoolIndex ");
    }

    s = RetrieveMessage(
            MSG_ANOMALY_POOLHEADER,
            FALSE,
            ItemInfo->Address,
            reason,
            ItemInfo->TypeSpecificParam
            );

    ASSERT(lstrlen(s) < EXPOOL_FORMAT_BUFFER_SIZE);
    lstrcpy(Buffer, s);
    FreeMessage(s);
}


VOID
EXPOOL_FormatAnomalyPoolDescriptor(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    LPWSTR s;

    s = RetrieveMessage(
            MSG_ANOMALY_POOLDESCRIPTOR_GENERAL,
            FALSE,
            ItemInfo->Address,
            ItemInfo->Size
            );
    ASSERT(lstrlen(s) < EXPOOL_FORMAT_BUFFER_SIZE);
    lstrcpy(Buffer, s);
    FreeMessage(s);
}


VOID
EXPOOL_FormatPoolDescriptor(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    UINT32 rc;
    LPWSTR s;
    POOL_DESCRIPTOR PoolDesc;
    SIZE_T BytesRead;
    WCHAR type_str[32];

    rc = KxReadMemory(h, NULL, ItemInfo->Address, &PoolDesc, sizeof(PoolDesc), &BytesRead);
    if (rc == NO_ERROR) {
        if (PoolDesc.PoolType == NonPagedPool) {
            lstrcpy(type_str, L"NonPagedPool");
        } else if (PoolDesc.PoolType == NonPagedPoolMustSucceed) {
            lstrcpy(type_str, L"NonPagedPoolMustSucceed");
        } else if (PoolDesc.PoolType == PagedPool) {
            lstrcpy(type_str, L"PagedPool");
        } else {
            lstrcpy(type_str, L"Unknown type");
        }

        s = RetrieveMessage(
                MSG_EXPOOL_POOLDESCRIPTOR,
                FALSE,
                ItemInfo->Address,
                ItemInfo->Size,
                type_str,
                PoolDesc.PoolIndex,
                PoolDesc.RunningAllocs,
                PoolDesc.RunningDeAllocs,
                PoolDesc.TotalPages,
                PoolDesc.TotalPages * PAGE_SIZE,
                PoolDesc.TotalBigPages,
                PoolDesc.TotalBigPages * PAGE_SIZE,
                PoolDesc.Threshold
                );

        ASSERT(lstrlen(s) < EXPOOL_FORMAT_BUFFER_SIZE);
        lstrcpy(Buffer, s);
        FreeMessage(s);
     } else {

        s = RetrieveMessage(
                MSG_EXPOOL_POOLDESCRIPTOR_UNREADABLE,
                FALSE,
                ItemInfo->Address,
                ItemInfo->Size
                );
        ASSERT(lstrlen(s) < EXPOOL_FORMAT_BUFFER_SIZE);
        lstrcpy(Buffer, s);
        FreeMessage(s);
     }
}



LPWSTR
EXPOOL_TagToString(
    ULONG32 PoolTag
    )
{
    ULONG32 TempTag = PoolTag;
    CHAR Tag[5];
    static WCHAR String[50];
    BOOL fPrintable = TRUE;
    int i;

#ifndef _NT351
    if (TempTag & PROTECTED_POOL) {
        TempTag &= ~PROTECTED_POOL;
    }
#endif //_NT351

    ZeroMemory(Tag, 5);
    CopyMemory(Tag, &TempTag, 4);

    for (i = 0; i < 4; i++) {
        if (!isprint((int)Tag[i])) fPrintable = FALSE;
    }

    ZeroMemory(String, 50);
    if (fPrintable) {
        wsprintf(String, L"%S", Tag);
    }

    return (LPWSTR)String;
}


UINT32
EXPOOL_PoolHeader_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    )
{
    UINT32 rc;
    KA_DATA_ITEM_INFO ItemInfo;
    SIZE_T BytesRead;
    POOL_HEADER PoolHeader;

    rc = KxQueryDataItem(h, ItemId, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    switch (MethodCode) {
    case M_POOL_HEAD_GET_BASIC_INFO:
        {
            //UINT32
            //(KANALYZE_CALL_METHOD_ROUTINE)(
            //    IN HKAPLUGIN PlugInHandle,
            //    IN KA_ITEM_ID ItemId,
            //    IN ULONG32 MethodCode,
            //    OUT PPOOL_HEAD_BASIC_INFO BasicInfo
            //    )
            PPOOL_HEAD_BASIC_INFO BasicInfo;
            POOL_TYPE PoolType;
            BasicInfo = va_arg(*Arguments, PPOOL_HEAD_BASIC_INFO);

            rc = KxReadMemory(h, NULL, ItemInfo.Address, &PoolHeader, sizeof(POOL_HEADER), &BytesRead);
            if (rc != NO_ERROR) {
                return rc;
            }

            PoolType = PoolHeader.PoolType;
            if (PoolType == 0) {
                BasicInfo->PoolType = KaFreedPool;
            } else {
                PoolType--;
                PoolType &= POOL_TYPE_MASK;
                switch (PoolType) {
                case NonPagedPool:
                    BasicInfo->PoolType = KaNonPagedPool;
                    break;
                case NonPagedPoolMustSucceed:
                    BasicInfo->PoolType = KaNonPagedPoolMustSucceed;
                    break;
                case PagedPool:
                    BasicInfo->PoolType = KaPagedPool;
                    break;
                default:
                    BasicInfo->PoolType = KaUnknownPool;
                }
            }

            BasicInfo->PoolIndex = (ULONG)PoolHeader.PoolIndex;
            BasicInfo->BlockNum = (ULONG)PoolHeader.BlockSize;
            BasicInfo->PreviousBlockNum = (ULONG)PoolHeader.PreviousSize;
            BasicInfo->PoolTag = (ULONG)PoolHeader.PoolTag;
            BasicInfo->ProtectedTag = FALSE;
#ifndef _NT351
            if (PoolHeader.PoolTag & PROTECTED_POOL) {
                BasicInfo->PoolTag &= ~PROTECTED_POOL;
                BasicInfo->ProtectedTag = TRUE;
            }
#endif //_NT351
            BasicInfo->Quota = (PoolHeader.PoolType & POOL_QUOTA_MASK) ? TRUE : FALSE;
            BasicInfo->QuotaProcess = BasicInfo->Quota ? PoolHeader.PoolTag : 0;

            return NO_ERROR;
        }
    default:
        //
        // unsupported method
        //
        return ERROR_INVALID_FUNCTION;
    }
}


UINT32
EXPOOL_PoolBlock_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    )
{
    UINT32 rc;
    KA_DATA_ITEM_INFO ItemInfo;

    rc = KxQueryDataItem(h, ItemId, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    switch (MethodCode) {
    case M_POOL_BLOCK_GET_BASIC_INFO:
        {
            //UINT32
            //(KANALYZE_CALL_METHOD_ROUTINE)(
            //    IN HKAPLUGIN PlugInHandle,
            //    IN KA_ITEM_ID ItemId,
            //    IN ULONG32 MethodCode,
            //    OUT PPOOL_BLOCK_BASIC_INFO BasicInfo
            //    )
            PPOOL_BLOCK_BASIC_INFO BasicInfo;
            BasicInfo = va_arg(*Arguments, PPOOL_BLOCK_BASIC_INFO);

            //
            // TypeSpecificParam of this item means an associated PoolBlock\PoolHeader's ID.
            // Zero means "No PoolHeader".
            //
            BasicInfo->PoolHead = ItemInfo.TypeSpecificParam;

            return NO_ERROR;
        }
    default:
        //
        // unsupported method
        //
        return ERROR_INVALID_FUNCTION;
    }
}



UINT32
EXPOOL_PoolDescriptor_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    )
{
    UINT32 rc;
    KA_DATA_ITEM_INFO ItemInfo;
    POOL_DESCRIPTOR PoolDesc;
    SIZE_T BytesRead;

    rc = KxQueryDataItem(h, ItemId, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    switch (MethodCode) {
    case M_POOL_DESCRIPTOR_GET_BASIC_INFO:
        {
            //UINT32
            //(KANALYZE_CALL_METHOD_ROUTINE)(
            //    IN HKAPLUGIN PlugInHandle,
            //    IN KA_ITEM_ID ItemId,
            //    IN ULONG32 MethodCode,
            //    OUT PPOOL_DESCRIPTOR_BASIC_INFO BasicInfo
            //    )
            PPOOL_DESCRIPTOR_BASIC_INFO BasicInfo;
            BasicInfo = va_arg(*Arguments, PPOOL_DESCRIPTOR_BASIC_INFO);

            rc = KxReadMemory(h, NULL, ItemInfo.Address, &PoolDesc, sizeof(POOL_DESCRIPTOR), &BytesRead);
            if (rc != NO_ERROR) {
                return rc;
            }

            switch (PoolDesc.PoolType & POOL_TYPE_MASK) {
            case NonPagedPool:
                BasicInfo->PoolType = KaNonPagedPool;
                break;
            case NonPagedPoolMustSucceed:
                BasicInfo->PoolType = KaNonPagedPoolMustSucceed;
                break;
            case PagedPool:
                BasicInfo->PoolType = KaPagedPool;
                break;
            default:
                BasicInfo->PoolType = KaUnknownPool;
            }
            BasicInfo->PoolIndex = PoolDesc.PoolIndex;
            BasicInfo->AllocCounts = PoolDesc.RunningAllocs;
            BasicInfo->FreeCounts = PoolDesc.RunningDeAllocs;
            BasicInfo->TotalAllocPages = PoolDesc.TotalPages;
            BasicInfo->TotalAllocBigPages = PoolDesc.TotalBigPages;
            BasicInfo->NumberOfFreeBlockIndex = POOL_LIST_HEADS;


            return NO_ERROR;
        }
    case M_POOL_DESCRIPTOR_GET_FREE_BLOCK_INFO:
        {
            //UINT32
            //(KANALYZE_CALL_METHOD_ROUTINE)(
            //    IN HKAPLUGIN PlugInHandle,
            //    IN KA_ITEM_ID ItemId,
            //    IN ULONG32 MethodCode,
            //    IN ULONG Index,
            //    OUT PPOOL_DESCRIPTOR_FREE_BLOCK_INFO FreeBlockInfo
            //    )
            ULONG Index;
            PPOOL_DESCRIPTOR_FREE_BLOCK_INFO FreeBlockInfo;
            KA_ITEM_ID TypeSpecificData[POOL_LIST_HEADS];

            Index = va_arg(*Arguments, ULONG);
            FreeBlockInfo = va_arg(*Arguments, PPOOL_DESCRIPTOR_FREE_BLOCK_INFO);

            if (Index >= POOL_LIST_HEADS || !FreeBlockInfo) {
                return ERROR_INVALID_PARAMETER;
            }

            if (ItemInfo.TypeSpecificDataSize == 0 || ItemInfo.TypeSpecificDataSize != sizeof(TypeSpecificData)) {
                return ERROR_INVALID_ACCESS;
            }

            rc = KxRetrieveTypeSpecificData(h, ItemId, sizeof(TypeSpecificData), TypeSpecificData, &BytesRead);
            if (rc != NO_ERROR) {
                return rc;
            }

            FreeBlockInfo->FreeListHead = TypeSpecificData[Index];

            return NO_ERROR;
        }
    default:
        //
        // unsupported method
        //
        return ERROR_INVALID_FUNCTION;
    }
}
