/*++

Copyright (c) 1998 Microsoft Corporation

Module Name:

    lookasid.c

Abstract:

    The plugin for checking and locating Lookaside List stuffs

Author:

    MSKK Oem Tools Project    07-Sep-1998

Revision History:

--*/

#include "precomp.h"
#pragma hdrstop


#ifndef _NT351

HKAPLUGIN hLOOKASIDE;

HKATYPE LOOKASIDE_Type_General = 0;            // LookasideList\\General
HKATYPE LOOKASIDE_Anomaly_General = 0;         // AnomalyDescriptor\\LookasideList\\General
HKATYPE LOOKASIDE_Type_SmallPool = 0;          // LookasideList\\SmallPool
HKATYPE LOOKASIDE_Anomaly_SmallPool = 0;       // AnomalyDescriptor\\LookasideList\\SmallPool

HKATYPE LOOKASIDE_Anomaly_InaccessibleMemory = 0; // AnomalyDescriptor\\InaccessibleMemory

#define LOOKASIDE_FORMAT_BUFFER_SIZE 1024

#define FLAG_PAGED_SMALL_POOL_LOOKASIDE 0x80000000

UINT_PTR
CALLBACK
LOOKASIDE_KanalyzeInterface(
    IN HKAPLUGIN PlugInHandle,
    IN UINT32 OperationCode,
    IN OUT UINT_PTR Param1,
    IN OUT PVOID Param2
    );

UINT32
LOOKASIDE_Initialize(
    IN HKAPLUGIN h,
    IN UINT_PTR Pass
    );

UINT32
LOOKASIDE_RegisterTypes(
    IN HKAPLUGIN h,
    IN PKANALYZE_REGISTER_TYPES_PROC fnRegister
    );

UINT32
LOOKASIDE_LocateItems(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fn
    );

UINT32
LOOKASIDE_ProcessItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fn
    );

UINT32
LOOKASIDE_NotifyItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fn
    );

UINT32
LOOKASIDE_Terminate(
    IN HKAPLUGIN h,
    IN KanalyzeFinalOutcome Outcome
    );

UINT32
LOOKASIDE_FormatDescription(
    IN HKAPLUGIN h,
    IN PKANALYZE_DATA_ITEM_STRING_REQUEST Request,
    OUT PUINT32 CharsWritten
    );

UINT32
LOOKASIDE_ProcessGeneralLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ThisId,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN PKANALYZE_LOCATE_PROC fnLocate
    );

UINT32
LOOKASIDE_LocateActiveLookasideAll(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    );

UINT32
LOOKASIDE_ProcessSmallPoolLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ThisId,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN PKANALYZE_LOCATE_PROC fnLocate
    );

UINT32
LOOKASIDE_LocateSmallPoolLookasideAll(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    );

VOID
LOOKASIDE_FormatGeneralLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
LOOKASIDE_FormatAnomalyGeneralLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
LOOKASIDE_FormatSmallPoolLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
LOOKASIDE_FormatAnomalySmallPoolLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

UINT32
LOOKASIDE_General_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    );

UINT32
LOOKASIDE_SmallPool_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    );

LPWSTR
LOOKASIDE_TagToString(
    ULONG32 PoolTag
    );



BOOL
LOOKASIDE_RegisterPlugIn(
    IN KANALYZE_REGISTER_PLUGIN_PROC RegistrationRoutine,
    IN ULONG32 CrashPlatform,
    IN ULONG32 CrashBuildNumber,
    PVOID Reserved1,
    UINT_PTR Reserved2
    )

/*++

Routine Description:


Arguments:


Return Value:


--*/

{

    hLOOKASIDE = (RegistrationRoutine)(
                     KANALYZE_INTERFACE_VERSION,
                     PLUGIN_LOOKASIDE,
                     LOOKASIDE_KanalyzeInterface,
                     &KanalyzeExtensionApis,
                     0,
                     0
                     );

    if (!hLOOKASIDE) {
        return FALSE;
    }

    return TRUE;
}



UINT_PTR
CALLBACK
LOOKASIDE_KanalyzeInterface(
    IN HKAPLUGIN PlugInHandle,
    IN UINT32 OperationCode,
    IN OUT UINT_PTR Param1,
    IN OUT PVOID Param2
    )

/*++

Routine Description:

    This is an interface routine between this plugin and the engine. The engine
    calls this routine.


Arguments:

    see the interface document

Return Value:

    Win32 Error Code

--*/

{
    UINT_PTR rc;

    rc = 0;

    switch (OperationCode) {
    case KA_INITIALIZE:
            rc = LOOKASIDE_Initialize(PlugInHandle, Param1);
            break;
    case KA_REGISTER_TYPES:
            rc = LOOKASIDE_RegisterTypes(PlugInHandle, (PKANALYZE_REGISTER_TYPES_PROC)Param2);
            break;
    case KA_START_LOCATE_ITEMS:
            rc = LOOKASIDE_LocateItems(PlugInHandle, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_PROCESS_ITEM:
            rc = LOOKASIDE_ProcessItem(PlugInHandle, (KA_ITEM_ID)Param1, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_NOTIFY_ITEM:
            rc = LOOKASIDE_NotifyItem(PlugInHandle, (KA_ITEM_ID)Param1, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_PROCESS_RESULT:
            rc = NO_ERROR; // not handle this code
            break;
    case KA_TERMINATE:
            rc = LOOKASIDE_Terminate(PlugInHandle, (KanalyzeFinalOutcome)Param1);
            break;
    case KA_FORMAT_DATA_ITEM_DESCRIPTION:
            rc = LOOKASIDE_FormatDescription(PlugInHandle, (PKANALYZE_DATA_ITEM_STRING_REQUEST)Param1, (PUINT32)Param2);
            break;
    default:
            rc = ERROR_INVALID_FUNCTION;
    }

    return rc;
}



UINT32
LOOKASIDE_Initialize(
    IN HKAPLUGIN h,
    IN UINT_PTR Pass
    )
{
    return NO_ERROR;
}



UINT32
LOOKASIDE_RegisterTypes(
    IN HKAPLUGIN h,
    IN PKANALYZE_REGISTER_TYPES_PROC fnRegister
    )
{
    UINT32 rc;

    rc = NO_ERROR;

    (fnRegister)(
        h,
        L"LookasideList\\General",
        L"Lookaside List Header",
        0,
        NULL,
        0,
        SupportFull,
        LOOKASIDE_General_Method,
        &LOOKASIDE_Type_General
        );

    (fnRegister)(
        h,
        L"AnomalyDescriptor\\LookasideList\\General",
        L"Anomaly Descriptor of LookasideList",
        0,
        NULL,
        KA_TC_DONT_CHECK_OVERLAP,
        SupportFull,
        NULL,
        &LOOKASIDE_Anomaly_General
        );

    (fnRegister)(
        h,
        L"LookasideList\\SmallPool",
        L"Lookaside List Header",
        0,
        NULL,
        0,
        SupportFull,
        LOOKASIDE_SmallPool_Method,
        &LOOKASIDE_Type_SmallPool
        );

    (fnRegister)(
        h,
        L"AnomalyDescriptor\\LookasideList\\SmallPool",
        L"Anomaly Descriptor of SmallPoolLookasideList",
        0,
        NULL,
        KA_TC_DONT_CHECK_OVERLAP,
        SupportFull,
        NULL,
        &LOOKASIDE_Anomaly_SmallPool
        );

    LOOKASIDE_Anomaly_InaccessibleMemory = KxGetTypeHandle(
                                               h,
                                               L"AnomalyDescriptor\\InaccessibleMemory"
                                               );

    return rc;
}


UINT32
LOOKASIDE_LocateItems(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;

    rc = LOOKASIDE_LocateActiveLookasideAll(h, fnLocate);
    if (rc != NO_ERROR) {
       return rc;
    }

    rc = LOOKASIDE_LocateSmallPoolLookasideAll(h, fnLocate);
    if (rc != NO_ERROR) {
       return rc;
    }

    return NO_ERROR;
}



UINT32
LOOKASIDE_ProcessItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    KA_DATA_ITEM_INFO ItemInfo;
    UINT32 rc;

    rc = KxQueryDataItem(h, id, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    if (ItemInfo.TypeHandle == LOOKASIDE_Type_General) {
        rc = LOOKASIDE_ProcessGeneralLookaside(h, id, &ItemInfo, fnLocate);
        return rc;
    }

    if (ItemInfo.TypeHandle == LOOKASIDE_Type_SmallPool) {
        rc = LOOKASIDE_ProcessSmallPoolLookaside(h, id, &ItemInfo, fnLocate);
        return rc;
    }

    return NO_ERROR;
}



UINT32
LOOKASIDE_NotifyItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    return NO_ERROR;
}



UINT32
LOOKASIDE_Terminate(
    IN HKAPLUGIN h,
    IN KanalyzeFinalOutcome Outcome
    )
{
    return NO_ERROR;
}



UINT32
LOOKASIDE_FormatDescription(
    IN HKAPLUGIN h,
    IN PKANALYZE_DATA_ITEM_STRING_REQUEST Request,
    OUT PUINT32 CharsWritten
    )
{
    UINT32 rc;
    WCHAR Buffer[LOOKASIDE_FORMAT_BUFFER_SIZE];
    KA_DATA_ITEM_INFO ItemInfo;
    UINT32 CharsNeeds;
    KA_ITEM_ID id;

    *CharsWritten = 0;

    id = Request->DataItemId;

    rc = KxQueryDataItem(h, id, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    if (ItemInfo.TypeHandle == LOOKASIDE_Type_General) {

        LOOKASIDE_FormatGeneralLookaside(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == LOOKASIDE_Anomaly_General) {

        LOOKASIDE_FormatAnomalyGeneralLookaside(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == LOOKASIDE_Type_SmallPool) {

        LOOKASIDE_FormatSmallPoolLookaside(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == LOOKASIDE_Anomaly_SmallPool) {

        LOOKASIDE_FormatAnomalySmallPoolLookaside(h, id, &ItemInfo, Buffer);

    } else {

        lstrcpy(Buffer, L"LOOKASIDE:\n");
    }

    CharsNeeds = lstrlen(Buffer) + 1;

    if (Request->MessageBufferSize < CharsNeeds) {
        *CharsWritten = CharsNeeds;
        return ERROR_MORE_DATA;
    } else {
        lstrcpy(Request->MessageBuffer, Buffer);
        *CharsWritten = CharsNeeds - 1;
        return NO_ERROR;
    }
}



///////////////////////////////////////////////////////////////////////////
UINT32
LOOKASIDE_ProcessGeneralLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ThisId,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;
    SIZE_T BytesRead;
    GENERAL_LOOKASIDE Lookaside;
    KANALYZE_DATA_ITEM DataItem;
    KA_ITEM_ID DataItemId;
    WCHAR ListHeadTypeName[MAX_TYPE_NAME];

    rc = KxReadMemory(h, NULL, ItemInfo->Address, &Lookaside, sizeof(GENERAL_LOOKASIDE), &BytesRead);
    if (rc != NO_ERROR) {
        //
        // unable to read memory, so locate special anomaly item here.
        //
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = LOOKASIDE_Anomaly_InaccessibleMemory;
        DataItem.Address = ItemInfo->Address;
        DataItem.Size = sizeof(GENERAL_LOOKASIDE);
        rc = (fnLocate)(h, ThisId, &DataItem, 0, &DataItemId);
    } else {
        //
        //BUGBUG validate insides here
        //


        //
        // locate single list header
        //
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        lstrcpy(ListHeadTypeName, L"SList\\SListHead\\GeneralLookaside");
        DataItem.TypeHandle = KxGetTypeHandle(h, ListHeadTypeName);
        DataItem.Address = ItemInfo->Address + FIELD_OFFSET(GENERAL_LOOKASIDE, ListHead);
        DataItem.Size = sizeof(SLIST_HEADER);
        rc = (fnLocate)(h, ThisId, &DataItem, 0, &DataItemId);

        //
        // BUGBUG validate each entries here
        //
    }

    return rc;
}



UINT32
LOOKASIDE_LocateActiveLookasideAll(
    HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;
    UINT_PTR HeadAddr, PrevAddr, NextAddr, Address;
    KANALYZE_DATA_ITEM DataItem;
    KA_ITEM_ID DataItemId;
    KA_ITEM_ID ListHeadId;
    LIST_ENTRY ListEntry;
    SIZE_T BytesRead;
    WCHAR ListHeadTypeName[MAX_TYPE_NAME];

    //
    // First, locate non-paged general lookaside list
    //
    HeadAddr = KxSymbolToAddress(h, L"ExNPagedLookasideListHead");
    ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
    lstrcpy(ListHeadTypeName, L"ListEntry\\ListHead\\ExNPagedLookasideList");
    DataItem.TypeHandle = KxGetTypeHandle(h, ListHeadTypeName);
    DataItem.Address = HeadAddr;
    DataItem.Size = sizeof(LIST_ENTRY);
    rc = (fnLocate)(h, 0, &DataItem, 0, &ListHeadId);
    if (rc != NO_ERROR) {
        return rc;
    }

    rc = KxReadMemory(h, NULL, HeadAddr, &ListEntry, sizeof(LIST_ENTRY), &BytesRead);
    if (rc != NO_ERROR) {
        MessageLog(h, LogWarnings, MSG_LOOKASIDE_CANNOT_READ_MEMORY, L"ExNPagedLookasideListHead", HeadAddr);
        return rc;
    }
    PrevAddr = HeadAddr;
    NextAddr = (UINT_PTR)ListEntry.Flink;
    while (NextAddr != HeadAddr) {
        rc = KxReadMemory(h, NULL, NextAddr, &ListEntry, sizeof(LIST_ENTRY), &BytesRead);
        if (rc != NO_ERROR || ((UINT_PTR)ListEntry.Blink != PrevAddr)) {
            //error
            MessageLog(h, LogWarnings, MSG_LOOKASIDE_INVALID_EXNPAGEDLOOKASIDELIST);
            goto locateactivelookasideall_skip1;
        }

        Address = (UINT_PTR)CONTAINING_RECORD(NextAddr, GENERAL_LOOKASIDE, ListEntry);

        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = LOOKASIDE_Type_General;
        DataItem.Address = Address;
        DataItem.Size = sizeof(NPAGED_LOOKASIDE_LIST);
        rc = (fnLocate)(h, ListHeadId, &DataItem, 0, &DataItemId);
        if (rc != NO_ERROR) {
            return rc;
        }

        PrevAddr = NextAddr;
        NextAddr = (UINT_PTR)ListEntry.Flink;
    }

locateactivelookasideall_skip1:

    //
    // Then, locate paged general lookaside list
    //
    HeadAddr = KxSymbolToAddress(h, L"ExPagedLookasideListHead");
    ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
    lstrcpy(ListHeadTypeName, L"ListEntry\\ListHead\\ExPagedLookasideList");
    DataItem.TypeHandle = KxGetTypeHandle(h, ListHeadTypeName);
    DataItem.Address = HeadAddr;
    DataItem.Size = sizeof(LIST_ENTRY);
    rc = (fnLocate)(h, 0, &DataItem, 0, &ListHeadId);
    if (rc != NO_ERROR) {
        return rc;
    }

    rc = KxReadMemory(h, NULL, HeadAddr, &ListEntry, sizeof(LIST_ENTRY), &BytesRead);
    if (rc != NO_ERROR) {
        MessageLog(h, LogWarnings, MSG_LOOKASIDE_CANNOT_READ_MEMORY, L"ExPagedLookasideListHead", HeadAddr);
        return rc;
    }
    PrevAddr = HeadAddr;
    NextAddr = (UINT_PTR)ListEntry.Flink;
    while (NextAddr != HeadAddr) {
        rc = KxReadMemory(h, NULL, NextAddr, &ListEntry, sizeof(LIST_ENTRY), &BytesRead);
        if (rc != NO_ERROR || ((UINT_PTR)ListEntry.Blink != PrevAddr)) {
            //error
            MessageLog(h, LogWarnings, MSG_LOOKASIDE_INVALID_EXPAGEDLOOKASIDELIST);
            goto locateactivelookasideall_skip2;
        }

        Address = (UINT_PTR)CONTAINING_RECORD(NextAddr, GENERAL_LOOKASIDE, ListEntry);

        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = LOOKASIDE_Type_General;
        DataItem.Address = Address;
        DataItem.Size = sizeof(PAGED_LOOKASIDE_LIST);
        rc = (fnLocate)(h, ListHeadId, &DataItem, 0, &DataItemId);
        if (rc != NO_ERROR) {
            return rc;
        }

        PrevAddr = NextAddr;
        NextAddr = (UINT_PTR)ListEntry.Flink;
    }

locateactivelookasideall_skip2:
    return rc;
}



UINT32
LOOKASIDE_ProcessSmallPoolLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ThisId,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;
    SIZE_T BytesRead;
    SMALL_POOL_LOOKASIDE Lookaside;
    KANALYZE_DATA_ITEM DataItem;
    KA_ITEM_ID DataItemId;
    WCHAR ListHeadTypeName[MAX_TYPE_NAME];

    KxReadMemory(h, NULL, ItemInfo->Address, &Lookaside, sizeof(SMALL_POOL_LOOKASIDE), &BytesRead); //never fail

    //
    // Locate single list header, connect it to this item.
    //
    ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
    lstrcpy(ListHeadTypeName, L"SList\\SListHead\\SmallPoolLookaside");
    DataItem.TypeHandle = KxGetTypeHandle(h, ListHeadTypeName);
    DataItem.Address = ItemInfo->Address + FIELD_OFFSET(SMALL_POOL_LOOKASIDE, SListHead);
    DataItem.Size = sizeof(SLIST_HEADER);
    rc = (fnLocate)(h, ThisId, &DataItem, 0, &DataItemId);
    if (rc != NO_ERROR) {
        return rc;
    }

    return rc;
}



UINT32
LOOKASIDE_LocateSmallPoolLookasideAll(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;
    UINT_PTR Address;
    KANALYZE_DATA_ITEM DataItem;
    KA_ITEM_ID SmallPoolId, PrevId;
    SIZE_T Size, BytesRead;
    ULONG i;
    SMALL_POOL_LOOKASIDE Lookaside[POOL_SMALL_LISTS];

    //
    // First, locate non-paged small pool lookaside list. Each small pool lookaside item
    // will be connected sequentially.
    //
    Address = KxSymbolToAddress(h, L"ExpSmallNPagedPoolLookasideLists");
    Size = sizeof(SMALL_POOL_LOOKASIDE) * POOL_SMALL_LISTS;
    rc = KxReadMemory(h, NULL, Address, &Lookaside, Size, &BytesRead);
    if (rc != NO_ERROR) {
        MessageLog(h, LogWarnings, MSG_LOOKASIDE_CANNOT_READ_MEMORY, L"ExpSmallNPagedPoolLookasideLists", Address);
        return rc;
    }

    PrevId = 0;
    for (i = 0; i < POOL_SMALL_LISTS; i++) {
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = LOOKASIDE_Type_SmallPool;
        DataItem.Address = Address + (sizeof(SMALL_POOL_LOOKASIDE) * i);
        DataItem.Size = sizeof(SMALL_POOL_LOOKASIDE);
        DataItem.TypeSpecificParam = i;
        rc = (fnLocate)(h, PrevId, &DataItem, 0, &SmallPoolId);
        if (rc != NO_ERROR) {
            return rc;
        }
        PrevId = SmallPoolId;
    }

    //
    // Next, locate paged small pool lookaside list. Each small pool lookaside item
    // will be connected sequentially.
    //
    Address = KxSymbolToAddress(h, L"ExpSmallPagedPoolLookasideLists");
    Size = sizeof(SMALL_POOL_LOOKASIDE) * POOL_SMALL_LISTS;
    rc = KxReadMemory(h, NULL, Address, &Lookaside, Size, &BytesRead);
    if (rc != NO_ERROR) {
        MessageLog(h, LogWarnings, MSG_LOOKASIDE_CANNOT_READ_MEMORY, L"ExpSmallPagedPoolLookasideLists", Address);
        return rc;
    }

    PrevId = 0;
    for (i = 0; i < POOL_SMALL_LISTS; i++) {
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = LOOKASIDE_Type_SmallPool;
        DataItem.Address = Address + (sizeof(SMALL_POOL_LOOKASIDE) * i);
        DataItem.Size = sizeof(SMALL_POOL_LOOKASIDE);
        DataItem.TypeSpecificParam = FLAG_PAGED_SMALL_POOL_LOOKASIDE | i;
        rc = (fnLocate)(h, PrevId, &DataItem, 0, &SmallPoolId);
        if (rc != NO_ERROR) {
            return rc;
        }
        PrevId = SmallPoolId;
    }

    return rc;
}



VOID
LOOKASIDE_FormatGeneralLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    UINT32 rc;
    WCHAR wbuff[MAX_PATH];
    SIZE_T Offset;
    LPWSTR s;
    LOOKASIDE_GENERAL_BASIC_INFO BasicInfo;
    LPWSTR TypeStr;
    LPWSTR wmsg;

    rc = KxGetSymbol(h, ItemInfo->Address, wbuff, MAX_PATH, &Offset);
    if (rc == NO_ERROR) {
        if (Offset != 0) wsprintf(&wbuff[lstrlen(wbuff)], L"+0x%x", Offset);
    } else {
        lstrcpy(wbuff, wmsg = RetrieveMessage(MSG_LOOKASIDE_NO_SYMBOL, FALSE));
        FreeMessage(wmsg);
    }

    rc = KxCallMethod(h, id, M_LOOKASIDE_GENERAL_BASIC_INFO, &BasicInfo);
    if (rc == NO_ERROR) {
        switch (BasicInfo.PoolType) {
        case KaNonPagedPool:
             TypeStr = RetrieveMessage(MSG_LOOKASIDE_NONPAGEDPOOL, FALSE);
             break;
        case KaNonPagedPoolMustSucceed:
             TypeStr = RetrieveMessage(MSG_LOOKASIDE_NONPAGEDMS, FALSE);
             break;
        case KaPagedPool:
             TypeStr = RetrieveMessage(MSG_LOOKASIDE_PAGEDPOOL, FALSE);
             break;
        default:
             TypeStr = RetrieveMessage(MSG_LOOKASIDE_UNKNOWN_TYPE, FALSE);
        }

        s = RetrieveMessage(
                MSG_LOOKASIDE_GENERAL_READABLE,
                FALSE,
                ItemInfo->Address,
                wbuff,
                TypeStr,
                LOOKASIDE_TagToString(BasicInfo.PoolTag),
                BasicInfo.BlockSize,
                BasicInfo.CurrentDepth,
                BasicInfo.MaximumDepth,
                BasicInfo.TotalAllocates,
                BasicInfo.AllocateMisses,
                BasicInfo.TotalFrees,
                BasicInfo.FreeMisses,
                BasicInfo.AllocateFunction,
                BasicInfo.FreeFunction
                );

        lstrcpy(Buffer, s);
        FreeMessage(s);
        FreeMessage(TypeStr);
    } else {
        s = RetrieveMessage(
                MSG_LOOKASIDE_GENERAL_UNREADABLE,
                FALSE,
                ItemInfo->Address,
                wbuff
                );
        lstrcpy(Buffer, s);
        FreeMessage(s);
    }
}


VOID
LOOKASIDE_FormatAnomalyGeneralLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    LPWSTR s;

    s = RetrieveMessage(MSG_ANOMALY_LOOKASIDE_GENERAL, FALSE, ItemInfo->Address);
    lstrcpy(Buffer, s);
    FreeMessage(s);
}



VOID
LOOKASIDE_FormatSmallPoolLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    UINT32 rc;
    WCHAR wbuff[MAX_PATH];
    SIZE_T Offset;
    LOOKASIDE_SMALLPOOL_BASIC_INFO BasicInfo;
    LPWSTR TypeStr;
    LPWSTR s;
    LPWSTR wmsg;

    rc = KxGetSymbol(h, ItemInfo->Address, wbuff, MAX_PATH, &Offset);
    if (rc == NO_ERROR) {
        if (Offset != 0) wsprintf(&wbuff[lstrlen(wbuff)], L"+0x%x", Offset);
    } else {
        lstrcpy(wbuff, wmsg = RetrieveMessage(MSG_LOOKASIDE_NO_SYMBOL, FALSE));
        FreeMessage(wmsg);
    }

    rc = KxCallMethod(h, id, M_LOOKASIDE_SMALLPOOL_BASIC_INFO, &BasicInfo);
    if (rc == NO_ERROR) {
        switch (BasicInfo.PoolType) {
        case KaNonPagedPool:
             TypeStr = RetrieveMessage(MSG_LOOKASIDE_NONPAGEDPOOL, FALSE);
             break;
        case KaNonPagedPoolMustSucceed:
             TypeStr = RetrieveMessage(MSG_LOOKASIDE_NONPAGEDMS, FALSE);
             break;
        case KaPagedPool:
             TypeStr = RetrieveMessage(MSG_LOOKASIDE_PAGEDPOOL, FALSE);
             break;
        default:
             TypeStr = RetrieveMessage(MSG_LOOKASIDE_UNKNOWN_TYPE, FALSE);
        }

        s = RetrieveMessage(
                MSG_LOOKASIDE_SMALLPOOL_READABLE,
                FALSE,
                ItemInfo->Address,
                wbuff,
                TypeStr,
                BasicInfo.CurrentDepth,
                BasicInfo.MaximumDepth,
                BasicInfo.TotalAllocates,
                BasicInfo.AllocateMisses,
                BasicInfo.TotalFrees,
                BasicInfo.FreeMisses
                );
        lstrcpy(Buffer, s);
        FreeMessage(s);
        FreeMessage(TypeStr);
    } else {
        s = RetrieveMessage(
                MSG_LOOKASIDE_SMALLPOOL_UNREADABLE,
                FALSE,
                ItemInfo->Address,
                wbuff
                );
        lstrcpy(Buffer, s);
        FreeMessage(s);
    }
}


VOID
LOOKASIDE_FormatAnomalySmallPoolLookaside(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    LPWSTR s;

    s = RetrieveMessage(MSG_ANOMALY_LOOKASIDE_SMALLPOOL, FALSE, ItemInfo->Address);
    lstrcpy(Buffer, s);
    FreeMessage(s);
}



BOOL
CALLBACK
LOOKASIDE_EnumSListHeadCallback(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID DataItemId,
    OUT PKA_ITEM_ID FoundId
    )
{
    *FoundId = DataItemId;
    return TRUE;
}



UINT32
LOOKASIDE_General_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    )
{
    UINT32 rc;
    KA_DATA_ITEM_INFO ItemInfo;
    GENERAL_LOOKASIDE Lookaside;
    SIZE_T BytesRead;
    KA_ENUM_STORED_ITEM_SPECIFICATION ItemSpec;
    KA_ITEM_ID SListHeadItemId;

    rc = KxQueryDataItem(h, ItemId, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    switch (MethodCode) {
    case M_LOOKASIDE_GENERAL_BASIC_INFO:
        {
            //UINT32
            //(KANALYZE_CALL_METHOD_ROUTINE)(
            //    IN HKAPLUGIN PlugInHandle,
            //    IN KA_ITEM_ID ItemId,
            //    IN ULONG32 MethodCode,
            //    OUT PLOOKASIDE_GENERAL_BASIC_INFO BasicInfo
            //    )
            PLOOKASIDE_GENERAL_BASIC_INFO BasicInfo;
            BasicInfo = va_arg(*Arguments, PLOOKASIDE_GENERAL_BASIC_INFO);

            rc = KxReadMemory(h, NULL, ItemInfo.Address, &Lookaside, sizeof(GENERAL_LOOKASIDE), &BytesRead);
            if (rc != NO_ERROR) {
                return rc;
            }

            switch (Lookaside.Type & POOL_TYPE_MASK) {
            case NonPagedPool:
                BasicInfo->PoolType = KaNonPagedPool;
                break;
            case NonPagedPoolMustSucceed:
                BasicInfo->PoolType = KaNonPagedPoolMustSucceed;
                break;
            case PagedPool:
                BasicInfo->PoolType = KaPagedPool;
                break;
            default:
                BasicInfo->PoolType = KaUnknownPool;
            }
            BasicInfo->PoolTag = Lookaside.Tag;
            BasicInfo->BlockSize = Lookaside.Size;
            BasicInfo->CurrentDepth = (ULONG)Lookaside.Depth;
            BasicInfo->MaximumDepth = (ULONG)Lookaside.MaximumDepth;
            BasicInfo->TotalAllocates = Lookaside.TotalAllocates;
            BasicInfo->AllocateMisses = Lookaside.AllocateMisses;
            BasicInfo->TotalFrees = Lookaside.TotalFrees;
            BasicInfo->FreeMisses = Lookaside.FreeMisses;
            BasicInfo->AllocateFunction = (UINT_PTR)Lookaside.Allocate;
            BasicInfo->FreeFunction = (UINT_PTR)Lookaside.Free;

            BasicInfo->SListHead = 0;

            ItemSpec.Flags = 0;
            ItemSpec.ReportedPlugIn = L"";
            ItemSpec.TypeSpecification = L"SList\\SListHead\\GeneralLookaside";
            ItemSpec.StartAddress = ItemInfo.Address + FIELD_OFFSET(GENERAL_LOOKASIDE, ListHead);
            ItemSpec.EndAddress = sizeof(SLIST_HEADER);
            ItemSpec.TypeSpecificParam = 0;
            rc = KxEnumStoredItem(
                     h,
                     &ItemSpec,
                     0,
                     0,
                     (PKANALYZE_ENUM_STORED_ITEM_CALLBACK)LOOKASIDE_EnumSListHeadCallback,
                     (PVOID)&SListHeadItemId
                     );
            if (rc == NO_ERROR) {
                BasicInfo->SListHead = SListHeadItemId;
            }

            return NO_ERROR;
        }
    default:
        //
        // unsupported method
        //
        return ERROR_INVALID_FUNCTION;
    }
}



UINT32
LOOKASIDE_SmallPool_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    )
{
    UINT32 rc;
    KA_DATA_ITEM_INFO ItemInfo;
    SMALL_POOL_LOOKASIDE Lookaside;
    SIZE_T BytesRead;
    KA_ENUM_STORED_ITEM_SPECIFICATION ItemSpec;
    KA_ITEM_ID SListHeadItemId;

    rc = KxQueryDataItem(h, ItemId, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    switch (MethodCode) {
    case M_LOOKASIDE_SMALLPOOL_BASIC_INFO:
        {
            //UINT32
            //(KANALYZE_CALL_METHOD_ROUTINE)(
            //    IN HKAPLUGIN PlugInHandle,
            //    IN KA_ITEM_ID ItemId,
            //    IN ULONG32 MethodCode,
            //    OUT PLOOKASIDE_SMALLPOOL_BASIC_INFO BasicInfo
            //    )
            PLOOKASIDE_SMALLPOOL_BASIC_INFO BasicInfo;
            BasicInfo = va_arg(*Arguments, PLOOKASIDE_SMALLPOOL_BASIC_INFO);

            rc = KxReadMemory(h, NULL, ItemInfo.Address, &Lookaside, sizeof(SMALL_POOL_LOOKASIDE), &BytesRead);
            if (rc != NO_ERROR) {
                return rc;
            }

            if (ItemInfo.TypeSpecificParam & FLAG_PAGED_SMALL_POOL_LOOKASIDE) {
                BasicInfo->PoolType = KaPagedPool;
            } else {
                BasicInfo->PoolType = KaNonPagedPool;
            }
            BasicInfo->BlockSize = ((ItemInfo.TypeSpecificParam & ~FLAG_PAGED_SMALL_POOL_LOOKASIDE)
                                   + 1) * POOL_SMALLEST_BLOCK;
            BasicInfo->CurrentDepth = (LONG)Lookaside.Depth;
            BasicInfo->MaximumDepth = (ULONG)Lookaside.MaximumDepth;
            BasicInfo->TotalAllocates = Lookaside.TotalAllocates;
            BasicInfo->AllocateMisses = Lookaside.TotalAllocates - Lookaside.AllocateHits;
            BasicInfo->TotalFrees = Lookaside.TotalFrees;
            BasicInfo->FreeMisses = Lookaside.TotalFrees - Lookaside.FreeHits;

            BasicInfo->SListHead = 0;

            ItemSpec.Flags = 0;
            ItemSpec.ReportedPlugIn = L"";
            ItemSpec.TypeSpecification = L"SList\\SListHead\\SmallPoolLookaside";
            ItemSpec.StartAddress = ItemInfo.Address + FIELD_OFFSET(GENERAL_LOOKASIDE, ListHead);
            ItemSpec.EndAddress = sizeof(SLIST_HEADER);
            ItemSpec.TypeSpecificParam = 0;
            rc = KxEnumStoredItem(
                     h,
                     &ItemSpec,
                     0,
                     0,
                     (PKANALYZE_ENUM_STORED_ITEM_CALLBACK)LOOKASIDE_EnumSListHeadCallback,
                     (PVOID)&SListHeadItemId
                     );
            if (rc == NO_ERROR) {
                BasicInfo->SListHead = SListHeadItemId;
            }

            return NO_ERROR;
        }
    default:
        //
        // unsupported method
        //
        return ERROR_INVALID_FUNCTION;
    }
}



LPWSTR
LOOKASIDE_TagToString(
    ULONG32 PoolTag
    )
{
    ULONG32 TempTag = PoolTag;
    CHAR Tag[5];
    static WCHAR LOOKASIDE_String[50];
    BOOL fPrintable = TRUE;
    int i;

#ifndef _NT351
    if (TempTag & PROTECTED_POOL) {
        TempTag &= ~PROTECTED_POOL;
    }
#endif //_NT351

    ZeroMemory(Tag, 5);
    CopyMemory(Tag, &TempTag, 4);

    for (i = 0; i < 4; i++) {
        if (!isprint((int)Tag[i])) fPrintable = FALSE;
    }

    ZeroMemory(LOOKASIDE_String, 50);
    if (fPrintable) {
        wsprintf(LOOKASIDE_String, L"%S", Tag);
    }

    return (LPWSTR)LOOKASIDE_String;
}



#endif //_NT351
