/*++

Copyright (c) 1998 Microsoft Corporation

Module Name:

    expool.c

Abstract:

    Hack Pool Enhancement Stuff

Author:

    MSKK Oem Tools Project    07-Sep-1998

Revision History:

--*/

#include "precomp.h"
#pragma hdrstop

HKAPLUGIN hPOOLEXT;

HKATYPE POOLEXT_Type_ExtraHeader = 0;        // PoolExt\\ExtraHeader
HKATYPE POOLEXT_Anomaly_ExtraHeader = 0;     // AnomalyDescriptor\\PoolExt\\ExtraHeader
HKATYPE POOLEXT_Type_PoolLog = 0;            // PoolExt\\PoolLog
HKATYPE POOLEXT_Type_EXPOOL_PoolHeader = 0;  // PoolBlock\\Header
HKATYPE POOLEXT_Anomaly_InaccessibleMemory = 0; // AnomalyDescriptor\\InaccessibleMemory

#define POOLEXT_FORMAT_BUFFER_SIZE 1024

//
// PoolExt internal data structure
//
typedef struct _PL_STATISTICS {
    SSIZE_T CurrentlyAllocated;
    SSIZE_T Overhead;
    ULONG32 AllocCount;
    ULONG32 FreeCount;
} PL_STATISTICS, *PPL_STATISTICS;

typedef struct _PL_TAG_ENTRY {
    ULONG32 Tag;
    ULONG32 Action;
} PL_TAG_ENTRY, *PPL_TAG_ENTRY;

typedef struct _PL_MODULE_ENTRY_V1 {
    PL_STATISTICS Statistics[2];
    PVOID DllBase;
    ULONG DllLimit;
    KSPIN_LOCK EntryLock;
    ULONG32 Action;
    ULONG32 TagCount;
    PPL_TAG_ENTRY TagEntry;
    WCHAR ModuleName[9];
    USHORT UnusedPadding;
} PL_MODULE_ENTRY_V1, *PPL_MODULE_ENTRY_V1;

typedef struct _PL_MODULE_ENTRY {
    PL_STATISTICS Statistics[2];
    PVOID DllBase;
    ULONG DllLimit;
    KSPIN_LOCK EntryLock;
    ULONG32 Action;
    ULONG32 TagCount;
    PPL_TAG_ENTRY TagEntry;
    WCHAR ModuleName[9];
    USHORT UnusedPadding;
    ULONG32 Pattern;                // added in phase2
    SIZE_T MaxFillLength;           // added in phase2
} PL_MODULE_ENTRY, *PPL_MODULE_ENTRY;

typedef struct _PL_TRACE_LOG {      // added in phase2
    ULONG32 TraceInfo;              // added in phase2
    PVOID BlockAddress;             // added in phase2
    SIZE_T BlockSize;               // added in phase2
    ULONG Tag;                      // added in phase2
} PL_TRACE_LOG, *PPL_TRACE_LOG;     // added in phase2

#define PL_TRACKING_MASK        64
#define PL_OVERHEAD_COUNT_MASK  32
#define PL_CHECKSUM_SHIFT       4
#define PL_LOW_BIT_MASK         2
#define PL_CHECKSUM_MASK        (1 << PL_CHECKSUM_SHIFT) - 1
#define PL_TRACKING_SPACE       sizeof(ULONG32)

#define STOPCODE_CAFE 0x0000CAFE

BOOL PlPresent = FALSE;
ULONG ModuleIndexShift;
ULONG ModuleIndexMask;
ULONG PlNumberOfEntry = 0;
PPL_MODULE_ENTRY PlModuleTable = NULL;

UINT_PTR PlLogBuffer;              // added in phase2
PPL_TRACE_LOG pPlLog;              // added in phase2
UINT_PTR PlLogBufferLimit;         // added in phase2
ULONG32 MaxLoggingEntry;           // added in phase2


BOOL POOLEXT_fBugcheckCafe = FALSE;
UINT_PTR POOLEXT_BugcheckPoolHeader = 0;

ULONG32 POOLEXT_Version;

UINT_PTR
CALLBACK
POOLEXT_KanalyzeInterface(
    IN HKAPLUGIN PlugInHandle,
    IN UINT32 OperationCode,
    IN OUT UINT_PTR Param1,
    IN OUT PVOID Param2
    );

UINT32
POOLEXT_Initialize(
    IN HKAPLUGIN h,
    IN UINT_PTR Pass
    );

UINT32
POOLEXT_RegisterTypes(
    IN HKAPLUGIN h,
    IN PKANALYZE_REGISTER_TYPES_PROC fnRegister
    );

UINT32
POOLEXT_LocateItems(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    );

UINT32
POOLEXT_ProcessItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fnLocate
    );

UINT32
POOLEXT_NotifyItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fnLocate
    );

UINT32
POOLEXT_Terminate(
    IN HKAPLUGIN h,
    IN KanalyzeFinalOutcome Outcome
    );

UINT32
POOLEXT_FormatDescription(
    IN HKAPLUGIN h,
    IN PKANALYZE_DATA_ITEM_STRING_REQUEST Request,
    OUT PUINT32 CharsWritten
    );

UINT32
POOLEXT_LocateExtraHeader(
    HKAPLUGIN h,
    KA_ITEM_ID id,
    PKANALYZE_LOCATE_PROC fnLocate,
    UINT_PTR AddressOfPoolHeader,
    SIZE_T SizeOfPoolHeader
    );

VOID
POOLEXT_FormatAnomalyExtraHeader(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
POOLEXT_FormatExtraHeader(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

VOID
POOLEXT_FormatPoolLog(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    );

UINT32
POOLEXT_ExtraHeader_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    );

BOOL
POOLEXT_SetupEnvironment(
    IN HKAPLUGIN h
    );

VOID
POOLEXT_FreeEnvironment(
    VOID
    );

BOOL
POOLEXT_CheckSumExtrabytes(
    IN ULONG32 Extrabytes
    );

ULONG32
POOLEXT_GetModuleIndexFromExtrabytes(
    IN ULONG32 Extrabytes
    );

SIZE_T
POOLEXT_GetCallerAddressFromExtrabytes(
    IN ULONG32 Extrabytes
    );

LPWSTR
POOLEXT_GetModuleNameFromModuleIndex(
    IN ULONG32 ModuleIndex
    );

UINT_PTR
POOLEXT_DllBaseFromModuleIndex(
    IN ULONG32 ModuleIndex
    );

LPWSTR
Allocator(
    IN HKAPLUGIN h,
    IN ULONG32 Extrabytes
    );



BOOL
POOLEXT_RegisterPlugIn(
    IN KANALYZE_REGISTER_PLUGIN_PROC RegistrationRoutine,
    IN ULONG32 CrashPlatform,
    IN ULONG32 CrashBuildNumber,
    PVOID Reserved1,
    UINT_PTR Reserved2
    )
{

    hPOOLEXT = (RegistrationRoutine)(
                   KANALYZE_INTERFACE_VERSION,
                   PLUGIN_POOLEXT,
                   POOLEXT_KanalyzeInterface,
                   &KanalyzeExtensionApis,
                   0,
                   0
                   );

    if (!hPOOLEXT) {
        return FALSE;
    }

    return TRUE;
}



UINT_PTR
CALLBACK
POOLEXT_KanalyzeInterface(
    IN HKAPLUGIN PlugInHandle,
    IN UINT32 OperationCode,
    IN OUT UINT_PTR Param1,
    IN OUT PVOID Param2
    )

/*++

Routine Description:

    This is an interface routine between this plugin and the engine. The engine
    calls this routine.


Arguments:

    see the interface document

Return Value:

    Win32 Error Code

--*/

{
    UINT_PTR rc;

    rc = 0;

    switch (OperationCode) {
    case KA_INITIALIZE:
            rc = POOLEXT_Initialize(PlugInHandle, Param1);
            break;
    case KA_REGISTER_TYPES:
            rc = POOLEXT_RegisterTypes(PlugInHandle, (PKANALYZE_REGISTER_TYPES_PROC)Param2);
            break;
    case KA_START_LOCATE_ITEMS:
            rc = POOLEXT_LocateItems(PlugInHandle, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_PROCESS_ITEM:
            rc = POOLEXT_ProcessItem(PlugInHandle, (KA_ITEM_ID)Param1, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_NOTIFY_ITEM:
            rc = POOLEXT_NotifyItem(PlugInHandle, (KA_ITEM_ID)Param1, (PKANALYZE_LOCATE_PROC)Param2);
            break;
    case KA_PROCESS_RESULT:
            rc = NO_ERROR; // not handle this code
            break;
    case KA_TERMINATE:
            rc = POOLEXT_Terminate(PlugInHandle, (KanalyzeFinalOutcome)Param1);
            break;
    case KA_FORMAT_DATA_ITEM_DESCRIPTION:
            rc = POOLEXT_FormatDescription(PlugInHandle, (PKANALYZE_DATA_ITEM_STRING_REQUEST)Param1, (PUINT32)Param2);
            break;
    default:
            rc = ERROR_INVALID_FUNCTION;
    }

    return rc;
}



UINT32
POOLEXT_Initialize(
    IN HKAPLUGIN h,
    IN UINT_PTR Pass
    )
{
    UINT32 rc;
    KANALYZE_CRASH_INFO CrashInfo;

    if (Pass == 0) {
        if (POOLEXT_SetupEnvironment(h)) {
            PlPresent = TRUE;
        }

        CrashInfo.SizeOfStruct = sizeof(KANALYZE_CRASH_INFO);
        rc = KxGetCrashInfo(h, &CrashInfo);
        if (rc != NO_ERROR) {
            return rc;
        }

        if (CrashInfo.BugcheckCode == STOPCODE_CAFE) {
            POOLEXT_fBugcheckCafe = TRUE;
            POOLEXT_BugcheckPoolHeader = CrashInfo.BugcheckParameters[1];
        }
    }

    return NO_ERROR;
}



UINT32
POOLEXT_RegisterTypes(
    IN HKAPLUGIN h,
    IN PKANALYZE_REGISTER_TYPES_PROC fnRegister
    )
{
    UINT32 rc;

    if (!PlPresent) {
        return NO_ERROR;
    }

    rc = (fnRegister)(
             h,
             L"PoolExt\\ExtraHeader",
             L"Pool Extra Header added by PoolExt Driver",
             0,
             NULL,
             0,
             SupportFull,
             POOLEXT_ExtraHeader_Method,
             &POOLEXT_Type_ExtraHeader
             );
    if (rc != NO_ERROR) {
        return rc;
    }

    rc = (fnRegister)(
             h,
             L"AnomalyDescriptor\\PoolExt\\ExtraHeader",
             L"Anomaly Descriptor for PoolExt\\ExtHeader",
             0,
             NULL,
             KA_TC_DONT_CHECK_OVERLAP,
             SupportFull,
             NULL,
             &POOLEXT_Anomaly_ExtraHeader
             );
    if (rc != NO_ERROR) {
        return rc;
    }

    rc = (fnRegister)(
             h,
             L"PoolExt\\PoolLog",
             L"Pool Enhancement Driver's Pool Log Buffer",
             0,
             NULL,
             0,
             SupportFull,
             NULL,
             &POOLEXT_Type_PoolLog
             );
    if (rc != NO_ERROR) {
        return rc;
    }


    //
    // demand notification of pool header here
    //
    rc = (fnRegister)(
             h,
             L"PoolBlock\\PoolHeader",
             NULL,
             0,
             NULL,
             0,
             SupportDemandNotification,
             NULL,
             &POOLEXT_Type_EXPOOL_PoolHeader
             );
    if (rc != NO_ERROR)
        return rc;

    POOLEXT_Anomaly_InaccessibleMemory = KxGetTypeHandle(
                                             h,
                                             L"AnomalyDescriptor\\InaccessibleMemory"
                                             );

    return rc;
}



UINT32
POOLEXT_LocateItems(
    IN HKAPLUGIN h,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;
    KANALYZE_DATA_ITEM DataItem;
    KA_ITEM_ID DataItemId;

    if (!PlPresent) {
        return NO_ERROR;
    }

    rc = NO_ERROR;

    if (POOLEXT_Version > 1 && PlLogBuffer != 0) {
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = POOLEXT_Type_PoolLog;
        DataItem.Address = PlLogBuffer;
        DataItem.Size = PlLogBufferLimit - PlLogBuffer;
        rc = (fnLocate)(h, 0, &DataItem, 0, &DataItemId);
    }

    return rc;
}



UINT32
POOLEXT_ProcessItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    if (!PlPresent) {
        return NO_ERROR;
    }

    return NO_ERROR;
}



UINT32
POOLEXT_NotifyItem(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKANALYZE_LOCATE_PROC fnLocate
    )
{
    UINT32 rc;
    KA_DATA_ITEM_INFO ItemInfo;

    if (!PlPresent) {
        return NO_ERROR;
    }

    rc = KxQueryDataItem(h, id, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    if (ItemInfo.TypeHandle != POOLEXT_Type_EXPOOL_PoolHeader) {
        // do nothing
        return NO_ERROR;
    }

    rc = POOLEXT_LocateExtraHeader(h, id, fnLocate, ItemInfo.Address, ItemInfo.Size);

    return rc;
}



UINT32
POOLEXT_Terminate(
    IN HKAPLUGIN h,
    IN KanalyzeFinalOutcome Outcome
    )
{
    if (!PlPresent) {
        return NO_ERROR;
    }

    POOLEXT_FreeEnvironment();

    return NO_ERROR;
}



UINT32
POOLEXT_FormatDescription(
    IN HKAPLUGIN h,
    IN PKANALYZE_DATA_ITEM_STRING_REQUEST Request,
    OUT PUINT32 CharsWritten
    )
{
    UINT32 rc;
    WCHAR Buffer[POOLEXT_FORMAT_BUFFER_SIZE];
    KA_DATA_ITEM_INFO ItemInfo;
    UINT32 CharsNeeds;
    KA_ITEM_ID id;

    *CharsWritten = 0;

    id = Request->DataItemId;

    rc = KxQueryDataItem(h, id, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    if (ItemInfo.TypeHandle == POOLEXT_Anomaly_ExtraHeader) {

        POOLEXT_FormatAnomalyExtraHeader(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == POOLEXT_Type_ExtraHeader) {

        POOLEXT_FormatExtraHeader(h, id, &ItemInfo, Buffer);

    } else if (ItemInfo.TypeHandle == POOLEXT_Type_PoolLog) {

        POOLEXT_FormatPoolLog(h, id, &ItemInfo, Buffer);

    } else {

        lstrcpy(Buffer, L"POOLEXT:\n");
    }

    CharsNeeds = lstrlen(Buffer) + 1;

    if (Request->MessageBufferSize < CharsNeeds) {
        *CharsWritten = CharsNeeds;
        return ERROR_MORE_DATA;
    } else {
        lstrcpy(Request->MessageBuffer, Buffer);
        *CharsWritten = CharsNeeds - 1;
        return NO_ERROR;
    }
}



UINT32
POOLEXT_LocateExtraHeader(
    HKAPLUGIN h,
    KA_ITEM_ID id,
    PKANALYZE_LOCATE_PROC fnLocate,
    UINT_PTR AddressOfPoolHeader,
    SIZE_T SizeOfPoolHeader
    )
{
    UINT32 rc;
    POOL_HEADER Header;
    UINT_PTR Address;
    SIZE_T BytesRead;
    ULONG32 ExtraBytes;
    KANALYZE_DATA_ITEM DataItem;
    KA_ITEM_ID ExtraItemId;

    rc = NO_ERROR;

    rc = KxReadMemory(h, NULL, AddressOfPoolHeader, &Header, sizeof(POOL_HEADER), &BytesRead);
    if (rc != NO_ERROR) {
        MessageLog(h, LogWarnings, MSG_POOLEXT_CANNOT_READ_MEMORY, L"Pool Header", AddressOfPoolHeader);
        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = POOLEXT_Anomaly_InaccessibleMemory;
        DataItem.Address = AddressOfPoolHeader;
        DataItem.Size = sizeof(POOL_HEADER);
        DataItem.TypeSpecificParam = 0;
        rc = (fnLocate)(h, id, &DataItem, 0, &ExtraItemId);
        return rc;
    }

    //
    // poolext driver sets private bit 7 in PoolHeader.PoolType in order to determine whether
    // this block is processed or not.
    //
    if ((Header.PoolType != 0 && (Header.PoolType - 1) & PL_TRACKING_MASK)
       || (POOLEXT_fBugcheckCafe && AddressOfPoolHeader == POOLEXT_BugcheckPoolHeader)) {
        ExtraBytes = 0;

        Address = AddressOfPoolHeader + ((Header.BlockSize) << POOL_BLOCK_SHIFT) - sizeof(ULONG32);

        if (Header.BlockSize == 0 ||
            PAGE_ALIGN(AddressOfPoolHeader) != PAGE_ALIGN(Address)) {

            //
            // this is illegal status, but don't process here. EXPOOL plug-in will process it.
            //
            return NO_ERROR;
        }

        ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
        DataItem.TypeHandle = POOLEXT_Type_ExtraHeader;
        DataItem.Address = Address;
        DataItem.Size = sizeof(ULONG32);
        DataItem.TypeSpecificParam = 0;
        rc = (fnLocate)(h, id, &DataItem, 0, &ExtraItemId);
        if (rc != NO_ERROR) {
            return rc;
        }

        rc = KxReadMemory(h, NULL, Address, &ExtraBytes, sizeof(ULONG32), &BytesRead);
        if (rc != NO_ERROR) {
            // never fail...
            return rc;
        }

        if (!POOLEXT_CheckSumExtrabytes(ExtraBytes)) {
            MessageLog(h, LogInformation, MSG_POOLEXT_FOUND_BROKEN_EXTRA_HEADER, Address, AddressOfPoolHeader);
            ZeroMemory((PVOID)&DataItem, sizeof(KANALYZE_DATA_ITEM));
            DataItem.TypeHandle = POOLEXT_Anomaly_ExtraHeader;
            DataItem.Address = Address;
            DataItem.Size = sizeof(ULONG32);
            DataItem.TypeSpecificParam = ANOMALY_POOLEXT_EH_CHECKSUM_MISMATCH;
            rc = (fnLocate)(h, ExtraItemId, &DataItem, 0, &ExtraItemId);
            if (rc != NO_ERROR) {
                return rc;
            }
        }
    }

    return rc;
}



VOID
POOLEXT_FormatAnomalyExtraHeader(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    LPWSTR s;

    s = RetrieveMessage(
            MSG_ANOMALY_BROKEN_EXTRA_HEADER,
            FALSE,
            ItemInfo->Address
            );

    ASSERT(lstrlen(s) < POOLEXT_FORMAT_BUFFER_SIZE);
    lstrcpy(Buffer, s);
    FreeMessage(s);
}



VOID
POOLEXT_FormatExtraHeader(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    UINT32 rc;
    LPWSTR s;
    SIZE_T BytesRead;
    ULONG32 ExtraBytes;

    rc = KxReadMemory(h, NULL, ItemInfo->Address, &ExtraBytes, sizeof(ULONG32), &BytesRead);
    if (rc != NO_ERROR) {
        s = RetrieveMessage(
                MSG_POOLEXT_EXTRAHEADER_UNREADABLE,
                FALSE,
                ItemInfo->Address
                );
        ASSERT(lstrlen(s) < POOLEXT_FORMAT_BUFFER_SIZE);
        lstrcpy(Buffer, s);
        FreeMessage(s);
        return;
    }

    s = RetrieveMessage(
            MSG_POOLEXT_EXTRAHEADER,
            FALSE,
            ItemInfo->Address,
            Allocator(h, ExtraBytes)
            );

    lstrcpy(Buffer, s);
    FreeMessage(s);
}



VOID
POOLEXT_FormatPoolLog(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID id,
    IN PKA_DATA_ITEM_INFO ItemInfo,
    IN OUT LPWSTR Buffer
    )
{
    LPWSTR s;

    s = RetrieveMessage(
            MSG_POOLEXT_POOLLOG,
            FALSE,
            ItemInfo->Address,
            ItemInfo->Address + ItemInfo->Size - 1
            );

    lstrcpy(Buffer, s);
    FreeMessage(s);
    return;
}



UINT32
POOLEXT_ExtraHeader_Method(
    IN HKAPLUGIN h,
    IN KA_ITEM_ID ItemId,
    IN ULONG32 MethodCode,
    IN va_list *Arguments
    )
{
    UINT32 rc;
    KA_DATA_ITEM_INFO ItemInfo;
    SIZE_T BytesRead;
    ULONG32 ExtraBytes;
    ULONG32 Index;
    LPWSTR ModuleName;

    rc = KxQueryDataItem(h, ItemId, 0, &ItemInfo);
    if (rc != NO_ERROR) {
        return rc;
    }

    switch (MethodCode) {
    case M_POOLEXT_EH_GET_ALLOCATOR_INFO:
        {
            //UINT32
            //(KANALYZE_CALL_METHOD_ROUTINE)(
            //    IN HKAPLUGIN PlugInHandle,
            //    IN KA_ITEM_ID ItemId,
            //    IN ULONG32 MethodCode,
            //    OUT PPOOLEXT_ALLOCATOR_INFO AllocatorInfo
            //    )
            PPOOLEXT_ALLOCATOR_INFO AllocatorInfo;
            AllocatorInfo = va_arg(*Arguments, PPOOLEXT_ALLOCATOR_INFO);

            rc = KxReadMemory(h, NULL, ItemInfo.Address, &ExtraBytes, sizeof(ULONG32), &BytesRead);
            if (rc != NO_ERROR) {
                return rc;
            }

            if (POOLEXT_CheckSumExtrabytes(ExtraBytes)) {
                //
                // checksum is good, so retrieve information
                //
                Index = POOLEXT_GetModuleIndexFromExtrabytes(ExtraBytes);
                ModuleName = POOLEXT_GetModuleNameFromModuleIndex(Index);
                if (ModuleName) {
                    lstrcpy(AllocatorInfo->AllocatorName, ModuleName);
                } else {
                    // never happen ?
                    AllocatorInfo->AllocatorName[0] = 0;
                }
                AllocatorInfo->CallerReturnAddress = POOLEXT_GetCallerAddressFromExtrabytes(ExtraBytes);

            } else {
                //
                // checsum is bad. we can't process any more.
                //
                AllocatorInfo->AllocatorName[0] = 0;
                AllocatorInfo->CallerReturnAddress = 0;
            }

            return NO_ERROR;
        }
    default:
        //
        // unsupported method
        //
        return ERROR_INVALID_FUNCTION;
    }
}





BOOL
POOLEXT_SetupEnvironment(
    IN HKAPLUGIN h
    )
{
    UINT32 rc;
    ULONG32 Size;
    SIZE_T BytesRead;
    UINT_PTR ModuleTableAddress;

    ModuleTableAddress = GetValue(h, L"Poolext!PlModuleTable");
    if (!ModuleTableAddress) {
        //
        // poolext.sys isn't present or no symbol for poolext.sys
        //
        MessageLog(h, LogInformation, MSG_POOEXT_NOT_PRESENT);
        return FALSE;
    }

    ModuleIndexShift = GetValue(h, L"Poolext!ModuleIndexShift");
    ModuleIndexMask = GetValue(h, L"Poolext!ModuleIndexMask");
    PlNumberOfEntry = GetValue(h, L"Poolext!PlNumberOfEntry");
    if (ModuleIndexShift < 6 || ModuleIndexShift > 9) {
        MessageLog(h, LogWarnings, MSG_POOLEXT_SYMBOL_NOT_MATCH);
        return FALSE;
    } else {
        if (((ULONG)(1 << ModuleIndexShift) - 1) != ModuleIndexMask) {
            MessageLog(h, LogWarnings, MSG_POOLEXT_SYMBOL_NOT_MATCH);
            return FALSE;
        }
    }

    PlLogBuffer = GetValue(h, L"Poolext!PlLogBuffer");
    PlLogBufferLimit = GetValue(h, L"Poolext!PlLogBufferLimit");
    MaxLoggingEntry = GetValue(h, L"Poolext!MaxLoggingEntry");
    if (!PlLogBuffer || !PlLogBufferLimit) {
        //
        // previous version of poolext.sys
        //
        MessageLog(h, LogInformation, MSG_POOLEXT_PREVIOUS_VERSION);
        POOLEXT_Version = 1;
    } else {
        POOLEXT_Version = 2;
    }


    Size = sizeof(PL_MODULE_ENTRY) * PlNumberOfEntry;
    PlModuleTable = (PPL_MODULE_ENTRY)LocalAlloc(LPTR, Size);
    if (!PlModuleTable) {
        MessageLog(h, LogWarnings, MSG_POOLEXT_ALLOCATE_MODULETABLE, Size);
        return FALSE;
    }

    if (POOLEXT_Version == 1) {
        PPL_MODULE_ENTRY_V1 ModuleTableV1;
        SIZE_T i;

        Size = sizeof(PL_MODULE_ENTRY_V1) * PlNumberOfEntry;
        ModuleTableV1 = (PPL_MODULE_ENTRY_V1)LocalAlloc(LPTR, Size);
        if (!ModuleTableV1) {
            MessageLog(h, LogWarnings, MSG_POOLEXT_ALLOCATE_MODULETABLE, Size);
            LocalFree(PlModuleTable);
            PlModuleTable = NULL;
            return FALSE;
        }

        rc = KxReadMemory(
                 h,
                 NULL,
                 ModuleTableAddress,
                 (PVOID)ModuleTableV1,
                 Size,
                 &BytesRead
                 );
        if (rc != NO_ERROR) {
            MessageLog(h, LogWarnings, MSG_POOLEXT_READ_MODULETABLE, ModuleTableAddress, Size);
            LocalFree(PlModuleTable);
            LocalFree(ModuleTableV1);
            PlModuleTable = NULL;
            return FALSE;
        }

        for (i = 0; i < PlNumberOfEntry; i++) {
            CopyMemory((PVOID)(&PlModuleTable[i]), (CONST VOID *)(&ModuleTableV1[i]), sizeof(PL_MODULE_ENTRY_V1));
        }

        LocalFree(ModuleTableV1);

        return TRUE;
    }

    rc = KxReadMemory(
             h,
             NULL,
             ModuleTableAddress,
             (PVOID)PlModuleTable,
             Size,
             &BytesRead
             );
    if (rc != NO_ERROR) {
        MessageLog(h, LogWarnings, MSG_POOLEXT_READ_MODULETABLE, ModuleTableAddress, Size);
        LocalFree(PlModuleTable);
        PlModuleTable = NULL;
        return FALSE;
    }

    return TRUE;
}



VOID
POOLEXT_FreeEnvironment(
    VOID
    )
{
    if (PlModuleTable) {
        LocalFree(PlModuleTable);
        PlModuleTable = NULL;
    }

    return;
}



BOOL
POOLEXT_CheckSumExtrabytes(
    IN ULONG32 Extrabytes
    )
{
    ULONG32 StoredChecksum;
    ULONG32 Checksum;
    int i;

    if (!Extrabytes)
        return FALSE;

    StoredChecksum = (Extrabytes >> ModuleIndexShift) & PL_CHECKSUM_MASK;
    Extrabytes &= ~(PL_CHECKSUM_MASK << ModuleIndexShift);  // disable checksum bits

    Checksum = 0;
    for (i = 0; i < 4; i++) {
        Checksum += ((Extrabytes >> (8 * i)) & 0xFF);
    }

    Checksum &= PL_CHECKSUM_MASK;

    if (StoredChecksum != Checksum) {
        return FALSE;
    }

    return TRUE;
}



ULONG32
POOLEXT_GetModuleIndexFromExtrabytes(
    IN ULONG32 Extrabytes
    )
{
    return (Extrabytes & ModuleIndexMask);
}



SIZE_T
POOLEXT_GetCallerAddressFromExtrabytes(
    IN ULONG32 Extrabytes
    )
{
    ULONG32 Index, Shift;
    UINT_PTR Address;
    SIZE_T Offset;

    Index = Extrabytes & ModuleIndexMask;

    //
    //    +----------------------------------+---+--------+
    //    |     CallerAddress / 4            |   |        |
    //    +----------------------------------+---+--------+
    //                  20bit                 4bit  8bit
    //

    Shift = ModuleIndexShift + PL_CHECKSUM_SHIFT;

    Offset = (Extrabytes >> Shift) << PL_LOW_BIT_MASK;

    //
    // if ModuleIndexShift is 8, Offset is 22 bit. We have to mask off 
    // the low 22 bit of DllBase.
    //

    Address = POOLEXT_DllBaseFromModuleIndex(Index);
    Address = Address >> (32 - Shift + PL_LOW_BIT_MASK);
    Address = Address << (32 - Shift + PL_LOW_BIT_MASK);

    Address |= Offset;

    return Address;
}



LPWSTR
POOLEXT_GetModuleNameFromModuleIndex(
    IN ULONG32 ModuleIndex
    )
{
    if (!PlPresent) return NULL;

    if (ModuleIndex < PlNumberOfEntry) {
        PlModuleTable[ModuleIndex].ModuleName[8] = UNICODE_NULL;  // cinch
        return &(PlModuleTable[ModuleIndex].ModuleName[0]);
    }

    return NULL;
}



UINT_PTR
POOLEXT_DllBaseFromModuleIndex(
    IN ULONG32 ModuleIndex
    )
{
    if (!PlPresent) return 0;

    if (ModuleIndex < PlNumberOfEntry) {
        return (UINT_PTR)(PlModuleTable[ModuleIndex].DllBase);
    }

    return 0;
}



LPWSTR
Allocator(
    IN HKAPLUGIN h,
    IN ULONG32 Extrabytes
    )
{
    UINT32 rc;
    LPWSTR ModuleName, s;
    ULONG32 ModuleIndex;
    SIZE_T CallerAddress;
    UINT_PTR DllBase;
    SSIZE_T Offset;
    WCHAR wbuff[MAX_PATH];
    static WCHAR POOLEXT_AllocatorNameBuffer[MAX_PATH];

    ZeroMemory(POOLEXT_AllocatorNameBuffer, sizeof(WCHAR) * MAX_PATH);
    ZeroMemory(wbuff, sizeof(WCHAR) * MAX_PATH);

    if (POOLEXT_CheckSumExtrabytes(Extrabytes)) {
        ModuleIndex = POOLEXT_GetModuleIndexFromExtrabytes(Extrabytes);
        CallerAddress = POOLEXT_GetCallerAddressFromExtrabytes(Extrabytes);

        ModuleName = POOLEXT_GetModuleNameFromModuleIndex(ModuleIndex);
        DllBase = POOLEXT_DllBaseFromModuleIndex(ModuleIndex);

        rc = KxGetSymbol(h, CallerAddress, wbuff, MAX_PATH, &Offset);
        if (rc == NO_ERROR) {
            //
            // add the offset to the end of string.
            //
            wsprintf(&wbuff[lstrlen(wbuff)], L"+%x", Offset);
        } else {
            wsprintf(wbuff, L"+%05x", CallerAddress - DllBase);
        }

        s = RetrieveMessage(
                MSG_POOLEXT_ALLOCATOR,
                FALSE,
                ModuleName,
                wbuff
                );
        if (s) {
            lstrcpy(POOLEXT_AllocatorNameBuffer, s);
            FreeMessage(s);
        }

    } else {
        s = RetrieveMessage(MSG_POOLEXT_CHECKSUM_MISMATCH, FALSE);
        if (s) {
            lstrcpy(POOLEXT_AllocatorNameBuffer, s);
            FreeMessage(s);
        }
    }

    return POOLEXT_AllocatorNameBuffer;
}



