/*++

Copyright (c) 1998 Microsoft Corporation

Module Name:

    utils.c

Abstract:

    Common Utilities Stuff

Author:

    MSKK Oem Tools Project    07-Sep-1998

Revision History:

--*/

#include "precomp.h"
#pragma hdrstop

extern HANDLE MyModuleHandle;

SIZE_T
GetValue (
    IN HKAPLUGIN h,
    IN LPCWSTR String
    )
{
    UINT32 rc;
    SIZE_T Location;
    SIZE_T Value;
    ULONG32 result;


    Location = KxSymbolToAddress(h, String);
    if (!Location) {
        return 0;
    }

    rc = KxReadMemory(h, NULL, Location, &Value, sizeof(SIZE_T), &result);
    if (rc != NO_ERROR || result < sizeof(SIZE_T)) {
        return 0;
    }

    return Value;
}



LPWSTR
RetrieveMessageV(
    IN UINT32 MessageId,
    IN BOOL SystemMessage,
    IN va_list *arglist
    )
{
    LPWSTR msg;
    DWORD rc;
    DWORD dwFlags;
    LPVOID lpSource;

    msg = NULL;

    dwFlags = SystemMessage ? FORMAT_MESSAGE_FROM_SYSTEM : FORMAT_MESSAGE_FROM_HMODULE;
    dwFlags |= FORMAT_MESSAGE_ALLOCATE_BUFFER;

    lpSource = SystemMessage ? NULL : MyModuleHandle;

    rc = FormatMessage(
             dwFlags,
             lpSource,
             MessageId,
             MAKELANGID(LANG_NEUTRAL, LANG_NEUTRAL),
             (LPWSTR)&msg,
             0,
             arglist
             );

    if (!rc) {
        return(NULL);
    }

    return msg;
}



LPWSTR
RetrieveMessage(
    IN UINT32 MessageId,
    IN BOOL SystemMessage,
    ...
    )
{
    va_list arglist;
    LPWSTR p;

    va_start(arglist, SystemMessage);
    p = RetrieveMessageV(MessageId, SystemMessage, &arglist);
    va_end(arglist);

    return p;
}



VOID
FreeMessage(
    IN LPWSTR String
    )
{
    LocalFree(String);
}



VOID
MessageLog(
    IN HKAPLUGIN h,
    IN KanalyzeLogLevel LogLevel,
    IN UINT32 MessageId,
    ...
    )
{
    va_list arglist;

    va_start(arglist, MessageId);
    KxWriteToLog(h, LogLevel, 0, NULL, MessageId, &arglist);
    va_end(arglist);
}


VOID
FormatLog(
    IN HKAPLUGIN h,
    IN KanalyzeLogLevel LogLevel,
    IN LPCWSTR FormatString,
    ...
    )
{
    va_list arglist;

    va_start(arglist, FormatString);
    KxWriteToLog(h, LogLevel, 0, FormatString, 0, &arglist);
    va_end(arglist);
}


